/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef REFERENCECOUNTED_HEADER
#define REFERENCECOUNTED_HEADER

#include <jmutex.h>
#include "common_irrlicht.h"

class ReferenceCounted
{
	u32 m_refcount;
	const char *m_name;
	JMutex m_refcount_mutex;

public:
	
	ReferenceCounted(const char *name=NULL):
		m_refcount(0), m_name(name)
	{
		m_refcount_mutex.Init();
	}
	
	void grab()
	{
		assert(m_refcount_mutex.IsInitialized());
		m_refcount_mutex.Lock();
		std::cout<<"ReferenceCounted::grab(): name="<<m_name<<" count="<<m_refcount<<"->"<<(m_refcount+1)<<std::endl;
		m_refcount++;
		// DEBUG
		/*if(m_refcount >= 3){
			std::cout<<"Refcount >= 3"<<std::endl;
		}*/
		m_refcount_mutex.Unlock();
	}

	void drop()
	{
		assert(m_refcount_mutex.IsInitialized());
		m_refcount_mutex.Lock();
		std::cout<<"ReferenceCounted::drop(): name="<<m_name<<" count="<<m_refcount<<"->"<<(m_refcount-1)<<std::endl;
		assert(m_refcount > 0);
		m_refcount--;
		m_refcount_mutex.Unlock();
	}

	u32 getRefcount()
	{
		return m_refcount;
	}
	
};

/*
	A reference to a ReferenceCounted object
*/
template <class T>
class Ref
{
	T * m_target;
public:
	Ref(T * target=NULL)
	{
		//assert(target != NULL);
		m_target = target;
		if(m_target != NULL)
			m_target->grab();
	}

	~Ref()
	{
		if(m_target != NULL)
			m_target->drop();
	}

	T * get()
	{
		return m_target;
	}
};

#endif

