/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAPSECTOR_HEADER
#define MAPSECTOR_HEADER

#include <jmutex.h>
#include "common_irrlicht.h"
#include "mapblock.h"
#include "referencecounted.h"
#include "heightmap.h"

/*
	This is an Y-wise stack of MapBlocks.
	
	TODO:
	- Add heightmap functionality
	- Implement node container functionality and modify Map to use it
*/


class MapSector : public ReferenceCounted, public HeightmapSource, public NodeContainer
{
	// The pile of MapBlocks
	core::map<s16, MapBlock*> m_blocks;
	//JMutex m_blocks_mutex; // For public access functions

	NodeContainer *m_parent;
	// Position on parent (in MapBlock widths)
	v2s16 m_pos;

	MapBlock *getBlockBuffered(s16 y);

public:

	JMutex mutex;

	MapSector(NodeContainer *parent, v2s16 pos):
			ReferenceCounted("MapSector"),
			m_parent(parent),
			m_pos(pos)
	{
		//m_blocks_mutex.Init();
		mutex.Init();
	}

	// If these return !NULL, the returned block should be
	// dropped() after use
	Ref<MapBlock> getBlock(s16 y);
	Ref<MapBlock> createBlankBlock(s16 y);
	
	core::list< Ref<MapBlock> > getBlocks();

	f32 getGroundHeight(v2s16 p);
	void setGroundHeight(v2s16 p, f32 h);
};

#endif

