/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef REFERENCECOUNTED_HEADER
#define REFERENCECOUNTED_HEADER

#include <jmutex.h>
#include "common_irrlicht.h"
#include <exception>

class NullPointerException : public std::exception
{
	virtual const char * what() const throw()
	{
		return "NullPointerException";
	}
};

class ReferenceCounted
{
public:
	
	class Exception : public std::exception
	{
		virtual const char * what() const throw()
		{
			return "ReferenceCounted::Exception";
		}
	};

	ReferenceCounted(const char *name=NULL):
		m_refcount(0), m_name(name)
	{
		m_refcount_mutex.Init();
	}
	
	void grab()
	{
		if(m_refcount_mutex.IsInitialized() == false)
			throw Exception();
		m_refcount_mutex.Lock();
		//std::cout<<"ReferenceCounted::grab(): name="<<m_name<<" count="<<m_refcount<<"->"<<(m_refcount+1)<<std::endl;
		m_refcount++;
		// DEBUG
		/*if(m_refcount >= 3){
			std::cout<<"Refcount >= 3"<<std::endl;
		}*/
		m_refcount_mutex.Unlock();
	}

	void drop()
	{
		if(m_refcount_mutex.IsInitialized() == false)
			throw Exception();
		m_refcount_mutex.Lock();
		//std::cout<<"ReferenceCounted::drop(): name="<<m_name<<" count="<<m_refcount<<"->"<<(m_refcount-1)<<std::endl;
		if(m_refcount == 0){
			m_refcount_mutex.Unlock();
			throw Exception();
		}
		m_refcount--;
		m_refcount_mutex.Unlock();
	}

	u32 getRefcount()
	{
		return m_refcount;
	}
	
private:

	u32 m_refcount;
	const char *m_name;
	JMutex m_refcount_mutex;

};

/*
	A reference to a ReferenceCounted object

	TODO: Does this need a mutex?

	TODO: Don't allow NULL pointers
*/
template <class T>
class Ref
{
	T * m_target;
public:
	Ref(T * target/*=NULL*/)
	{
		// Debug stuff
		//std::cout<<"Ref::Ref(target="<<(int)target<<")"<<std::endl;

		//assert(target != NULL);

		if(target == NULL)
			throw NullPointerException();
		
		m_target = target;
		//if(m_target != NULL)
		m_target->grab();
	}

	Ref(const Ref &ref)
	{
		// Debug stuff
		//std::cout<<"Ref::Ref(ref&.get()="<<(int)ref.get()<<")"<<std::endl;

		//assert(target != NULL);
		
		m_target = ref.get();
		//if(m_target != NULL)
		m_target->grab();
	}

	Ref & operator = (const Ref &ref)
	{
		// Debug stuff
		/*if(ref.get() == NULL){
			std::cout<<"Ref::operator=(Ref&): ref.get()=NULL"<<std::endl;
		}
		else{
			std::cout<<"Ref::operator=(Ref&): ref.get()->getRefcount()="<<ref.get()->getRefcount()<<std::endl;
		}*/

		m_target = ref.get();
		//if(m_target != NULL)
		m_target->grab();
			
		return *this;
	}
	
	//TODO: remove?
	Ref & operator = (const T *target)
	{
		//std::cout<<"Ref::operator=(T*): target="<<(int)target<<std::endl;
		//assert(0);

		if(target == NULL)
			throw NullPointerException();
		
		m_target = target;
		//if(m_target != NULL)
		m_target->grab();
	}

	~Ref()
	{
		// Debug stuff
		//std::cout<<"Ref::~Ref(m_target="<<(int)m_target<<")"<<std::endl;

		dropTarget();
	}

	T * get() const
	{
		return m_target;
	}

	void dropTarget()
	{
		//if(m_target != NULL){
		m_target->drop();
		// Cannot be referenced anymore
		//m_target = NULL;
		//}
	}
};

#endif

