#ifndef SOCKET_HEADER
#define SOCKET_HEADER

#ifdef _WIN32
	#include <winsock2.h>
	#pragma comment(lib, "wsock32.lib")
typedef SOCKET socket_t;
typedef int socklen_t;
#else
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <fcntl.h>
	#include <unistd.h>
typedef int socket_t;
#endif

#include "exceptions.h"

class SocketException : public BaseException
{
public:
	SocketException(const char *s):
		BaseException(s)
	{
	}
};

void sockets_init();
void sockets_cleanup();

class Address
{
public:
	Address();
	Address(unsigned int address, unsigned short port);
	Address(unsigned int a, unsigned int b,
			unsigned int c, unsigned int d,
			unsigned short port);
	unsigned int getAddress() const;
	unsigned short getPort() const;
	void print() const;
private:
	unsigned int m_address;
	unsigned short m_port;
};

class UDPSocket
{
public:
	UDPSocket();
	~UDPSocket();
	void Bind(unsigned short port);
	//void Close();
	//bool IsOpen();
	void Send(const Address & destination, const void * data, int size);
	int Receive(Address & sender, void * data, int size);
	int GetHandle(); // For debugging purposes only
private:
	int m_handle;
};

#endif

