#include "socket.h"
#include <stdio.h>
#include <iostream>

// Debug print options
#define DP 0
#define DPS "    "

bool g_sockets_initialized = false;

void sockets_init()
{
#ifdef _WIN32
	WSADATA WsaData;
	if(WSAStartup( MAKEWORD(2,2), &WsaData ) != NO_ERROR)
		throw SocketException("WSAStartup failed");
#else
#endif
	g_sockets_initialized = true;
}

void sockets_cleanup()
{
#ifdef _WIN32
	WSACleanup();
#endif
}

Address::Address()
{
}

Address::Address(unsigned int address, unsigned short port)
{
	m_address = address;
	m_port = port;
}

Address::Address(unsigned int a, unsigned int b,
		unsigned int c, unsigned int d,
		unsigned short port)
{
	m_address = (a<<24) | (b<<16) | ( c<<8) | d;
	m_port = port;
}

unsigned int Address::getAddress() const
{
	return m_address;
}

unsigned short Address::getPort() const
{
	return m_port;
}

void Address::print() const
{
	std::cout<<((m_address>>24)&0xff)<<"."
			<<((m_address>>16)&0xff)<<"."
			<<((m_address>>8)&0xff)<<"."
			<<((m_address>>0)&0xff)<<":"
			<<m_port;
}

UDPSocket::UDPSocket()
{
	if(g_sockets_initialized == false)
		throw SocketException("Sockets not initialized");
	
    m_handle = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	
	if(DP)
	std::cout<<DPS<<"UDPSocket("<<(int)m_handle<<")::UDPSocket()"<<std::endl;
	
    if(m_handle <= 0)
    {
		throw SocketException("Failed to create socket");
    }
}

UDPSocket::~UDPSocket()
{
	if(DP)
	std::cout<<DPS<<"UDPSocket("<<(int)m_handle<<")::~UDPSocket()"<<std::endl;

#ifdef _WIN32
	closesocket(m_handle);
#else
	close(m_handle);
#endif
}

void UDPSocket::Bind(unsigned short port)
{
	if(DP)
	std::cout<<DPS<<"UDPSocket("<<(int)m_handle<<")::Bind(): port="<<port<<std::endl;

    sockaddr_in address;
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(port);

    if(bind(m_handle, (const sockaddr*)&address, sizeof(sockaddr_in)) < 0)
    {
		throw SocketException("Failed to bind socket");
    }
	
#ifdef _WIN32
	DWORD nonblocking = 1;
	if(ioctlsocket(handle, FIONBIO, &nonblocking) != 0)
	{
		throw SocketException("Failed set non-blocking mode");
	}
#else
	int nonblocking = 1;
	if(fcntl(m_handle, F_SETFL, O_NONBLOCK, nonblocking) == -1)
	{
		throw SocketException("Failed set non-blocking mode");
	}
#endif
}

void UDPSocket::Send(const Address & destination, const void * data, int size)
{
	if(DP){
		std::cout<<DPS<<"UDPSocket("<<(int)m_handle
				<<")::Send(): destination=";
		destination.print();
		//std::cout<<", size="<<size<<std::endl;
		std::cout<<", size="<<size<<", data=";
		for(int i=0; i<size; i++){
			if(i%2==0) printf(" ");
			printf("%.2X", ((int)((const char*)data)[i])&0xff);
		}
		std::cout<<std::endl;
	}

	sockaddr_in address;
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(destination.getAddress());
	address.sin_port = htons(destination.getPort());

	int sent = sendto(m_handle, (const char*)data, size,
		0, (sockaddr*)&address, sizeof(sockaddr_in));

    if(sent != size)
    {
		throw SocketException("Failed to send packet");
    }
}

int UDPSocket::Receive(Address & sender, void * data, int size)
{
	sockaddr_in address;
	socklen_t address_len = sizeof(address);

	int received = recvfrom(m_handle, (char*)data,
			size, 0, (sockaddr*)&address, &address_len);

	if(received <= 0)
		return 0;

	unsigned int address_ip = ntohl(address.sin_addr.s_addr);
	unsigned int address_port = ntohs(address.sin_port);

	sender = Address(address_ip, address_port);

	if(DP){
		std::cout<<DPS<<"UDPSocket("<<(int)m_handle<<")::Receive(): sender=";
		sender.print();
		//std::cout<<", received="<<received<<std::endl;
		std::cout<<", received="<<received<<", data=";
		for(int i=0; i<received; i++){
			if(i%2==0) printf(" ");
			printf("%.2X", ((int)((const char*)data)[i])&0xff);
		}
		std::cout<<std::endl;
	}

	return received;
}

int UDPSocket::GetHandle()
{
	return m_handle;
}


