#include "connection.h"

namespace con
{

/*void makePacket(u8 *packetdata, u8 *data, u32 datasize,
		u32 protocol_id, u16 peer_id, u16 seqnum, u8 type)*/
BufferedPacket makePacket(Address &address, u8 *data, u32 datasize,
		u32 protocol_id, u16 peer_id, u16 seqnum, u8 type)
{
	u32 packet_size = datasize + HEADER_SIZE;
	BufferedPacket p(packet_size);
	p.address = address;

	writeU32(&p.data[0], protocol_id);
	writeU16(&p.data[4], peer_id);
	writeU16(&p.data[6], seqnum);
	writeU8(&p.data[8], type);

	memcpy(&p.data[HEADER_SIZE], data, datasize);

	return p;
}

Connection::Connection(u32 protocol_id)
{
	m_protocol_id = protocol_id;
	m_peer_id = PEER_ID_NEW;
	m_waiting_new_peer_id = false;
}

Connection::~Connection()
{
	// Clear peers
	core::map<u16, Peer*>::Iterator j;
	j = m_peers.getIterator();
	for(; j.atEnd() == false; j++)
	{
		Peer *peer = j.getNode()->getValue();
		delete peer;
	}
}

void Connection::Serve(unsigned short port)
{
	m_socket.Bind(port);
	m_peer_id = PEER_ID_SERVER;
}

void Connection::Connect(Address address)
{
	core::map<u16, Peer*>::Node *node = m_peers.find(PEER_ID_SERVER);
	if(node != NULL){
		throw ConnectionException("Already connected to a server");
	}

	Peer *peer = new Peer();
	peer->m_address = address;
	peer->m_id = PEER_ID_SERVER;
	m_peers.insert(peer->m_id, peer);
	
	m_socket.Bind(0);

	// Send a dummy packet to server with peer_id = PEER_ID_NEW
	m_peer_id = PEER_ID_NEW;
	Send(PEER_ID_SERVER, NULL, 0, true);
	//ProtocolSend(address, NULL, 0, PEER_ID_NEW, 

	m_waiting_new_peer_id = true;
}

bool Connection::Connected()
{
	if(m_peers.size() != 1)
		return false;
		
	core::map<u16, Peer*>::Node *node = m_peers.find(PEER_ID_SERVER);
	if(node == NULL)
		return false;
	
	if(m_peer_id == PEER_ID_NEW)
		return false;
	
	return true;
}

u32 Connection::Receive(u16 &peer_id, u8 *data, s32 datasize, u16 &seqnumber)
{
	for(;;)
	{
		s32 headersize = HEADER_SIZE;
		/*
			Check if there is a suitable packet in the reliable
			packet queue of some peer
		*/
		core::map<u16, Peer*>::Iterator j;
		j = m_peers.getIterator();
		for(; j.atEnd() == false; j++)
		{
			Peer *peer = j.getNode()->getValue();
			assert(j.getNode()->getKey() == peer->m_id);
			
			core::list<BufferedPacket>::Iterator i;
check_first_on_queue:
			// Get first item of buffer
			i = peer->m_incoming_reliable_queue.begin();
			// If queue is empty, continue to next peer
			if(i == peer->m_incoming_reliable_queue.end()){
				//std::cout<<"Queue is empty"<<std::endl;
				// Queue is empty
				continue;
			}
		
			PrintInfo();
			std::cout<<"Receive(): Processing queued packets for peer_id="
					<<peer->m_id<<"... ";

			std::cout<<"next seqnum should be "
					<<peer->m_next_incoming_reliable_seqnum
					<<"; ";

			// Read the sequence number of the first one
			u16 firstseqnum = readSeqNumber(*i->data);
			std::cout<<"First seqnum is "<<firstseqnum<<"; ";

			// If it is old, dump it and start again
			// TODO: seqnum wrap-around
			if(firstseqnum < peer->m_next_incoming_reliable_seqnum)
			{
				std::cout<<"First one is old, removing; ";
				//delete[] i->data;
				peer->m_incoming_reliable_queue.erase(i);
				goto check_first_on_queue;
			}
			// If it has the right sequence number, remove it from
			// the queue and return the packet.
			// TODO: seqnum wrap-around
			else if(firstseqnum ==
					peer->m_next_incoming_reliable_seqnum)
			{
				std::cout<<"Taking the first one from queue"<<std::endl;

				peer->m_next_incoming_reliable_seqnum++;

				assert(datasize >= (s32)i->data.getSize() - headersize);
				u32 received_datasize = i->data.getSize() - headersize;
				memcpy(data, &i->data[headersize], received_datasize);

				peer_id = readPeerId(*i->data);
				seqnumber = readSeqNumber(*i->data);

				//delete[] i->data;
				peer->m_incoming_reliable_queue.erase(i);
				
				return received_datasize;
			}
			else{
				std::cout<<"The first one in queue comes later"<<std::endl;
			}
		}

		/*
			Receive a packet from the network
		*/

		s32 packet_maxsize = datasize + headersize;
		//u8 packetdata[packet_maxsize];
		//u8 *packetdata = new u8[packet_maxsize];
		Buffer<u8> packetdata(packet_maxsize);

		Address sender;

		s32 received_size = m_socket.Receive(sender, *packetdata, packet_maxsize);
	
		//std::cout<<"received_size="<<received_size<<std::endl;
		if(received_size <= 0)
			throw NoIncomingDataException("No incoming data");
		if(received_size < headersize)
			throw InvalidIncomingDataException("No full header received");
		if(readU32(&packetdata[0]) != m_protocol_id)
			throw InvalidIncomingDataException("Invalid protocol id");
		
		peer_id = readPeerId(*packetdata);
		seqnumber = readSeqNumber(*packetdata);
		u8 type = readType(*packetdata);

		if(peer_id == PEER_ID_NEW){
			// Somebody wants to make a new connection

			// Get a unique peer id (2 or higher)
			u16 peer_id_new = 2;
			core::map<u16, Peer*>::Iterator j;
			j = m_peers.getIterator();
			for(; j.atEnd() == false; j++)
			{
				Peer *peer = j.getNode()->getValue();
				if(peer_id_new <= peer->m_id)
					peer_id_new = peer->m_id + 1;
			}

			PrintInfo();
			std::cout<<"Receive(): Got a PEER_ID_NEW, giving peer_id="
					<<peer_id_new<<std::endl;

			// Create a peer
			Peer *peer = new Peer();
			peer->m_address = sender;
			peer->m_id = peer_id_new;
			m_peers.insert(peer->m_id, peer);

			// Create packet contents
			u8 data[2];
			writeU16(&data[0], peer_id_new);
			// Seqnum, type
			u16 seqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 type = TYPE_PEER_ID;
			// Send it
			ProtocolSend(sender, data, 2, m_peer_id, seqnum, type);
		
			// We're now talking to a valid peer_id
			peer_id = peer_id_new;
			
			/*// Process more packets
			delete packetdata;
			continue;*/
		}

		/*PrintInfo();
		std::cout<<"peer_id="<<peer_id<<std::endl;*/
		
		core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

		if(node == NULL){
			// Peer not found
			PrintInfo();
			std::cout<<"Receive(): Peer not found"<<std::endl;
			throw InvalidIncomingDataException("Peer not found (possible timeout)");
		}

		Peer *peer = node->getValue();

		//TODO: Validate peer address
		
		// Update port
		//peer->m_address.setPort(sender.getPort());
		//peer->m_address.setAddress(sender.getAddress());

		peer->m_timeout_counter = 0.0;
		
		if(type == TYPE_PING)
		{
			// Send PONG
			u16 packetseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 packettype = TYPE_PONG;
			ProtocolSend(sender, NULL, 0, m_peer_id, packetseqnum, packettype);
		}
		else if(type == TYPE_PONG)
		{
		}
		else if(type == TYPE_ACK)
		{
			if(received_size < headersize + 2)
				throw InvalidIncomingDataException("Short data");
			/*
				The packet is an ACK.
				Remove corresponding packet from ack timeout queue
			*/

			u16 ackseqnum = readU16(&packetdata[headersize]);

			/*PrintInfo();
			std::cout<<"Receive(): Got an ACK packet for seqnum="
					<<ackseqnum<<" from peer_id="
					<<peer_id<<std::endl;*/
			
			//std::cout<<"Going through ack_timeout_queue"<<std::endl;

			core::list<BufferedPacket>::Iterator i;
			i = peer->m_ack_timeout_queue.begin();
			for(; i != peer->m_ack_timeout_queue.end(); i++)
			{
				u16 seqnum = readSeqNumber(*i->data);
				//std::cout<<"seqnum="<<seqnum<<std::endl;
				if(seqnum == ackseqnum){
					PrintInfo();
					std::cout<<"Receive(): Found ACKed packet in buffer"
							". Removing it."<<std::endl;
					
					// Decrease resend timeout
					peer->m_resend_timeout /= 2.0;
					if(peer->m_resend_timeout < RESEND_TIMEOUT_MINIMUM)
						peer->m_resend_timeout = RESEND_TIMEOUT_MINIMUM;

					// Found packet in buffer
					// delete it
					//delete[] i->data;
					peer->m_ack_timeout_queue.erase(i);
					break;
				}
			}

			// Process more packets
			//delete packetdata;
			continue;
		}
		else if(type == TYPE_PEER_ID)
		{
			if(received_size < headersize + 2)
				throw InvalidIncomingDataException("Short data");
			/*
				We don't allow changing our peer id if we alredy have one
				Also, happens in this (harmless) glitch:
				- Client sends reliable packet with peer_id=0
				- Server sends TYPE_ACK
				- Server sends TYPE_PEER_ID
				- Client gets TYPE_PEER_ID
				- Client doesn't get ACK, sends a peer_id=0 packet again
				- Client gets TYPE_PEER_ID again -> goes here
			*/
			if(m_peer_id != PEER_ID_NEW){
				throw InvalidIncomingDataException(
						"Got a PEER_ID packet but we already have one");
			}

			u16 new_peer_id = readU16(&packetdata[headersize]);
			PrintInfo();
			std::cout<<"Receive(): Changing peer id to "<<new_peer_id<<std::endl;
			m_peer_id = new_peer_id;

			// Process more packets
			//delete packetdata;
			continue;
		}
		else if(type == TYPE_RELIABLE)
		{
			/*
				The packet is a reliable packet:
				- Send an ACK
				- If the packet is the one expected next, return it.
				- If the packet is to be delivered later, queue it and get a
				  new packet.
			*/

			PrintInfo();
			std::cout<<"Receive(): Received a reliable packet, seqnumber="
					<<seqnumber<<", sending ACK."<<std::endl;
			
			// Send ACK
			u8 ackdata[2];
			writeU16(&ackdata[0], seqnumber);
			u16 ackseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 acktype = TYPE_ACK;
			ProtocolSend(sender, ackdata, 2, m_peer_id, ackseqnum, acktype);
			
			// If it is an old packet, dump it.
			// TODO: seqnum wrap-around
			if(seqnumber < peer->m_next_incoming_reliable_seqnum)
			{
				PrintInfo();
				std::cout<<"Receive(): Packet is old"
						<<" (seqnumber="<<seqnumber<<")"<<std::endl;
				// Process more packets
				//delete packetdata;
				continue;
			}
			
			// Deliver the packet if it is the one expected next
			if(seqnumber == peer->m_next_incoming_reliable_seqnum)
			{
				PrintInfo();
				std::cout<<"Receive(): Packet has the right seqnum"<<std::endl;
				// It will get returned if we do nothing here.
				// The next one is going to be one higher.
				peer->m_next_incoming_reliable_seqnum++;
			}
			// If the packet is not the one expected next, queue it
			// and get a new one.
			else
			{
				PrintInfo();
				std::cout<<"Receive(): Queueing incoming packet "
						<<"(expected seqnum="
						<<peer->m_next_incoming_reliable_seqnum
						<<", got seqnum="<<seqnumber<<")"
						<<std::endl;

				// Make packet with copied data
				BufferedPacket p(received_size);
				p.time = 0.0;
				//p.data = new u8[received_size];
				//p.size = received_size;
				memcpy(*p.data, *packetdata, received_size);

				// Find the appropriate place in the receive queue
				core::list<BufferedPacket>::Iterator i;
				i = peer->m_incoming_reliable_queue.begin();
				bool added = false;
				// If queue is empty
				if(i == peer->m_incoming_reliable_queue.end())
				{
					PrintInfo();
					std::cout<<"Receive(): Queue is empty, adding to"
							" beginning"<<std::endl;
					peer->m_incoming_reliable_queue.push_front(p);
					added = true;
				}
				/* If the sequence number of the first packet in queue is
				   higher than that of the packet we got, add the packet
				   to the beginning of the list */
				else if(readSeqNumber(*i->data) > seqnumber){
					PrintInfo();
					std::cout<<"Receive(): First has higher number,"
							" adding to beginning"
							<<std::endl;
					peer->m_incoming_reliable_queue.push_front(p);
					added = true;
				}
				/*
					Else find the right place and add it there.
				*/
				else{
					u16 count = 1;
					for(;;){
						if(readSeqNumber(*i->data) < seqnumber){
							PrintInfo();
							std::cout<<"Receive(): Adding after index "<<count
									<<std::endl;
							peer->m_incoming_reliable_queue.insert_after(i, p);
							added = true;
							break;
						}
						i++;
						count++;
					}
				}
				assert(added == true);

				// Process more packets
				continue;
			}
		}
		else if(type == TYPE_SPLIT)
		{
			if(received_size < headersize + 4)
				throw InvalidIncomingDataException("Short data");
			/*PrintInfo();
			std::cout<<"Receive(): Received a split packet chunk, seqnumber="
					<<seqnumber<<", sending ACK."<<std::endl;*/
			
			u16 chunkcount = readU16(&packetdata[headersize+0]);
			u16 chunknum = readU16(&packetdata[headersize+2]);

			// Send ACK
			u8 ackdata[4];
			writeU16(&ackdata[0], seqnumber);
			writeU16(&ackdata[2], chunknum);
			u16 ackseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 acktype = TYPE_SPLIT_ACK;
			ProtocolSend(sender, ackdata, 4, m_peer_id, ackseqnum, acktype);
			
			// Add chunk to buffer

			IncomingSplitPacket *splitpacket;
			
			// If no chunks of this packet have arrived, add it to buffer
			if(peer->m_incoming_split_queue.find(seqnumber) == NULL)
			{
				IncomingSplitPacket *newsplitpacket
						= new IncomingSplitPacket();
				peer->m_incoming_split_queue[seqnumber] = newsplitpacket;
			}
			
			splitpacket = peer->m_incoming_split_queue[seqnumber];
			
			// Ignore chunk if it has already been added
			if(splitpacket->chunks.find(chunknum) == NULL)
			{
				u32 chunkdatasize = received_size - headersize - 4;
				SharedBuffer<u8> chunkdata(chunkdatasize);
				memcpy(*chunkdata, &packetdata[headersize+4], chunkdatasize);

				splitpacket->chunks[chunknum] = chunkdata;
				splitpacket->chunk_count = chunkcount;


				// If all chunks are now received, handle the result
				if(splitpacket->allReceived())
				{
					PrintInfo();
					std::cout<<"Receive(): A complete split packet received."
							" seqnumber="<<seqnumber
							<<", chunkcount="<<chunkcount
							<<std::endl;
					/*
						Construct the original data to the data buffer
					*/

					// Calculate total size
					u32 totalsize = 0;
					core::map<u16, SharedBuffer<u8> >::Iterator i;
					i = splitpacket->chunks.getIterator();
					for(; i.atEnd() == false; i++)
					{
						totalsize += i.getNode()->getValue().getSize();
					}

					// Check that the given data buffer is big enough
					if(datasize < (s32)totalsize){
						// Remove splitpacket from buffer
						peer->m_incoming_split_queue.remove(seqnumber);
						delete splitpacket;
						// Throw exception
						throw InvalidIncomingDataException
								("Split packet is too big for given buffer");
					}

					// Copy chunks to data buffer
					u32 start = 0;
					for(u32 chunk_i=0; chunk_i<splitpacket->chunk_count;
							chunk_i++)
					{
						SharedBuffer<u8> buf = splitpacket->chunks[chunk_i];
						u16 chunkdatasize = buf.getSize();
						memcpy(&data[start], *buf, chunkdatasize);
						start += chunkdatasize;;
					}

					// Remove splitpacket from buffer
					peer->m_incoming_split_queue.remove(seqnumber);
					delete splitpacket;
					
					// Return resulting data as received
					peer_id = peer_id;
					seqnumber = 0;

					return totalsize;
				}
			}
			
			// No full packet yet
			// Process more packets
			continue;
		}
		else if(type == TYPE_SPLIT_ACK)
		{
			if(received_size < headersize + 4)
				throw InvalidIncomingDataException("Short data");
				
			u16 ackseqnum = readU16(&packetdata[headersize]);
			u16 ackchunknum = readU16(&packetdata[headersize+2]);
			core::list<BufferedPacket>::Iterator i;
			i = peer->m_split_ack_timeout_queue.begin();
			for(; i != peer->m_split_ack_timeout_queue.end(); i++)
			{
				u16 seqnum = readSeqNumber(*i->data);
				u16 chunknum = readU16(&i->data[headersize+2]);
				//std::cout<<"seqnum="<<seqnum<<std::endl;
				if(seqnum == ackseqnum && chunknum == ackchunknum){
					/*PrintInfo();
					std::cout<<"Receive(): Found ACKed split packet"
							" chunk in buffer. Removing it."<<std::endl;*/
					
					// Decrease resend timeout
					peer->m_resend_timeout /= 2.0;
					if(peer->m_resend_timeout < RESEND_TIMEOUT_MINIMUM)
						peer->m_resend_timeout = RESEND_TIMEOUT_MINIMUM;

					// Found packet in buffer -> erase it
					peer->m_split_ack_timeout_queue.erase(i);
					break;
				}
			}
			// Process more packets
			continue;
		}

		assert(datasize >= received_size - headersize);
		u32 received_datasize = received_size - headersize;
		memcpy(data, &packetdata[headersize], received_datasize);

		return received_datasize;
	}
}

void Connection::SendToAll(u8 *data, u32 datasize, bool reliable)
{
	core::map<u16, Peer*>::Iterator j;
	j = m_peers.getIterator();
	for(; j.atEnd() == false; j++)
	{
		Peer *peer = j.getNode()->getValue();
		Send(peer->m_id, data, datasize, reliable);
	}
}

void Connection::Send(u16 peer_id, u8 *data, u32 datasize, bool reliable)
{
	PrintInfo();
	std::cout<<"Send(): peer_id="<<peer_id<<", datasize="<<datasize
			<<", reliable="<<reliable<<std::endl;

	core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

	if(node == NULL){
		// Peer not found
		throw InvalidOutgoingDataException("Peer not found (possible timeout)");
	}

	Peer *peer = node->getValue();
	
	// A reality check
	assert(peer->m_id == peer_id);

	u8 type = TYPE_NORMAL;
	u16 seqnum;
	if(reliable){
		type = TYPE_RELIABLE;
		seqnum = peer->m_next_seqnum_reliable;
		peer->m_next_seqnum_reliable++;
	}
	else{
		seqnum = peer->m_next_seqnum_unreliable;
		peer->m_next_seqnum_unreliable++;
	}

	/*u32 packet_size = datasize + HEADER_SIZE;
	BufferedPacket b(packet_size);
	b.address = peer->m_address;
	makePacket(*b.data, data, datasize, m_protocol_id,
			m_peer_id, seqnum, type);*/
	BufferedPacket b = makePacket(peer->m_address, data, datasize,
			m_protocol_id, m_peer_id, seqnum, type);
	
	/*
		If it is a reliable packet, add it to m_ack_timeout_queue
	*/
	if(reliable == true)
	{
		peer->m_ack_timeout_queue.push_back(b);
	}
	
	RawSend(b);
	//RawSend(b.address, *b.data, b.data.getSize());
	//ProtocolSend(destination, data, datasize, m_peer_id, seqnum, type);
}

void Connection::SendSplit(u16 peer_id, u8 *data, u32 datasize)
{
	/*
		TODO:
		Split packet in parts and send them reliably to the other end,
		without using the connection-wide reliability system.
		The packets should be sized at 512 bytes maximum.
		
		Split packet seqnum is unrelated to unreliable and reliable
		seqnums. It represents the unique id of the packet.

		Split packet format:
		- u16 chunk_count
		- u16 chunk_num
		- data
	*/
	
	Peer *peer = GetPeer(peer_id);

	u16 seqnum = peer->m_next_seqnum_split;
	peer->m_next_seqnum_split++;

	core::list<SharedBuffer<u8> > chunkdatas;

	u32 maximum_data_size = 512 - HEADER_SIZE - 2 - 2;
	u32 start = 0;
	u32 end = 0;
	u32 chunk_num = 0;
	do{
		end = start + maximum_data_size - 1;
		if(end > datasize - 1)
			end = datasize - 1;
		
		u32 payload_size = end - start + 1;
		u32 packet_size = 2 + 2 + payload_size;

		//BufferedPacket chunk(packet_size);
		SharedBuffer<u8> chunkdata(packet_size);

		// [0]=chunk_count is written at next stage
		writeU16(&chunkdata[2], chunk_num);
		memcpy(&chunkdata[4], &data[start], payload_size);

		chunkdatas.push_back(chunkdata);
		
		start = end + 1;
		chunk_num++;
	}
	while(end != datasize - 1);

	u16 chunk_count = chunkdatas.getSize();

	core::list<SharedBuffer<u8> >::Iterator i = chunkdatas.begin();
	for(; i != chunkdatas.end(); i++)
	{
		u8 *chunkdata = **i;
		//u32 packet_size = i->getSize() + HEADER_SIZE;
		
		// Write chunk_count
		writeU16(&chunkdata[0], chunk_count);
		
		/*BufferedPacket packet(packet_size);
		packet.address = peer->m_address;
		makePacket(*packet.data, chunkdata, datasize, m_protocol_id,
				m_peer_id, seqnum, TYPE_SPLIT);*/
				
		BufferedPacket packet = makePacket(peer->m_address, chunkdata,
				i->getSize(), m_protocol_id, m_peer_id, seqnum, TYPE_SPLIT);
		
		// Add to ack timeout queue so that we can check the ACK.
		peer->m_split_ack_timeout_queue.push_back(packet);
		
		RawSend(packet);
		/*RawSend(packet.address, *packet.data,
				packet.data.getSize());*/
	}
}

void Connection::ProtocolSend(Address &destination, u8 *data, u32 datasize,
		u16 peer_id, u16 seqnum, u8 type)
{
	BufferedPacket packet = makePacket(destination, data, datasize,
			m_protocol_id, peer_id, seqnum, type);

	RawSend(packet);
}

//void Connection::RawSend(Address &destination, u8 *data, u32 datasize)
void Connection::RawSend(BufferedPacket &packet)
{
	//m_socket.Send(destination, data, datasize);
	m_socket.Send(packet.address, *packet.data, packet.data.getSize());
}

void Connection::RunTimeouts(float dtime)
{
	/*
		1) Remove timed-out peers
		2) Send timed-out packets in m_ack_timeout_queue
	*/
	core::list<u16> timeouted_peers;
	core::map<u16, Peer*>::Iterator j;
	j = m_peers.getIterator();
	for(; j.atEnd() == false; j++)
	{
		Peer *peer = j.getNode()->getValue();

		u16 peer_id = j.getNode()->getKey();
		assert(peer_id == peer->m_id);

		// Check peer timeout
		peer->m_timeout_counter += dtime;
		// TODO: Get the constant timeout from somewhere else
		if(peer->m_timeout_counter > 10.0)
		{
			PrintInfo();
			std::cout<<"RunTimeouts(): Peer "<<peer_id
					<<" has timeouted."<<std::endl;
			// Add peer to the list
			timeouted_peers.push_back(peer_id);
			// Don't bother going through the buffers of this one
			continue;
		}

		// Ping peer at constant intervals
		peer->m_ping_timer += dtime;
		// TODO: Get the constant interval from somewhere else
		if(peer->m_ping_timer > 3.0)
		{
			// Send PING
			u16 packetseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 packettype = TYPE_PING;
			ProtocolSend(peer->m_address, NULL, 0, m_peer_id,
					packetseqnum, packettype);
			peer->m_ping_timer = 0.0;
		}

		// Check reliable packet buffer timeouts
		core::list<BufferedPacket>::Iterator i;
		i = peer->m_ack_timeout_queue.begin();
		for(; i != peer->m_ack_timeout_queue.end(); i++)
		{
			i->time += dtime;
			// If it has run over timeout, re-send it
			// TODO: Get the constant timeout from somewhere else
			if(i->time > peer->m_resend_timeout)
			{
				PrintInfo();
				std::cout<<"RunTimeouts(): Re-sent packet."<<std::endl;
				//RawSend(i->address, *i->data, i->data.getSize());
				RawSend(*i);
				i->time = 0.0;

				peer->m_resend_timeout *= 2.0;
				if(peer->m_resend_timeout > RESEND_TIMEOUT_MAXIMUM)
					peer->m_resend_timeout = RESEND_TIMEOUT_MAXIMUM;
			}
		}

		// Check split packet buffer timeouts
		i = peer->m_split_ack_timeout_queue.begin();
		for(; i != peer->m_split_ack_timeout_queue.end(); i++)
		{
			i->time += dtime;
			// If it has run over timeout, re-send it
			// TODO: Get the constant timeout from somewhere else
			if(i->time > peer->m_resend_timeout)
			{
				PrintInfo();
				std::cout<<"RunTimeouts(): Re-sent split packet chunk."
						<<std::endl;
				RawSend(*i);
				i->time = 0.0;

				peer->m_resend_timeout *= 2.0;
				if(peer->m_resend_timeout > RESEND_TIMEOUT_MAXIMUM)
					peer->m_resend_timeout = RESEND_TIMEOUT_MAXIMUM;
			}
		}
	}
	// Remove timeouted peers
	core::list<u16>::Iterator i = timeouted_peers.begin();
	for(; i != timeouted_peers.end(); i++)
	{
		PrintInfo();
		std::cout<<"RunTimeouts(): Removing peer "<<(*i)<<std::endl;
		m_peers.remove(*i);
	}
}

Peer* Connection::GetPeer(u16 peer_id)
{
	core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

	if(node == NULL){
		// Peer not found
		throw NotFoundException("Peer not found (possible timeout)");
	}

	return node->getValue();
}

void Connection::PrintInfo()
{
	std::cout<<"Connection(h="<<m_socket.GetHandle()<<", id="
			<<m_peer_id<<"): ";
}

} // namespace

