/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAP_HEADER
#define MAP_HEADER

#include <jmutex.h>
#include <jthread.h>
#include <iostream>
#include <malloc.h>

#ifdef _WIN32
	#include <windows.h>
	#define sleep_s(x) Sleep((x*1000))
#else
	#include <unistd.h>
	#define sleep_s(x) sleep(x)
#endif

#include "common_irrlicht.h"
#include "heightmap.h"
#include "loadstatus.h"
#include "mapnode.h"
#include "mapblock.h"
#include "mapsector.h"
#include "constants.h"

class InvalidFilenameException : public BaseException
{
public:
	InvalidFilenameException(const char *s):
		BaseException(s)
	{}
};

/*class Map;

class MapUpdateThread : public JThread
{
	bool run;
	JMutex run_mutex;

	Map *map;

public:

	MapUpdateThread(Map *the_map) : JThread(), run(true), map(the_map)
	{
		run_mutex.Init();
	}

	void * Thread();

	bool getRun()
	{
		run_mutex.Lock();
		bool run_cached = run;
		run_mutex.Unlock();
		return run_cached;
	}
	void setRun(bool a_run)
	{
		run_mutex.Lock();
		run = a_run;
		run_mutex.Unlock();
	}
};*/

class Map : public NodeContainer, public Heightmappish
{
protected:

	std::ostream &m_dout;

	core::map<v2s16, MapSector*> m_sectors;
	JMutex m_getsector_mutex;
	JMutex m_gensector_mutex;

	v3f camera_position;
	v3f camera_direction;
	JMutex camera_mutex;

	//MapUpdateThread updater;

	// Be sure to set this to NULL when the cached sector is deleted 
	MapSector *m_sector_cache;
	v2s16 m_sector_cache_p;

	WrapperHeightmap m_hwrapper;

public:

	v3s16 drawoffset;

	//LoadStatus status;
	
	Map(std::ostream &dout);
	virtual ~Map();

	void updateCamera(v3f pos, v3f dir)
	{
		camera_mutex.Lock();
		camera_position = pos;
		camera_direction = dir;
		camera_mutex.Unlock();
	}

	/*void StartUpdater()
	{
		updater.Start();
	}

	void StopUpdater()
	{
		updater.setRun(false);
		while(updater.IsRunning())
			sleep_s(1);
	}

	bool UpdaterIsRunning()
	{
		return updater.IsRunning();
	}*/

	/*
		Returns integer position of the node in given
		floating point position.
	*/
	static v3s16 floatToInt(v3f p)
	{
		v3s16 p2(
			(p.X + (p.X>0 ? BS/2 : -BS/2))/BS,
			(p.Y + (p.Y>0 ? BS/2 : -BS/2))/BS,
			(p.Z + (p.Z>0 ? BS/2 : -BS/2))/BS);
		return p2;
	}

	static v3f intToFloat(v3s16 p)
	{
		v3f p2(
			p.X * BS,
			p.Y * BS,
			p.Z * BS
		);
		return p2;
	}

	static core::aabbox3d<f32> getNodeBox(v3s16 p)
	{
		return core::aabbox3d<f32>(
			(float)p.X * BS - 0.5*BS,
			(float)p.Y * BS - 0.5*BS,
			(float)p.Z * BS - 0.5*BS,
			(float)p.X * BS + 0.5*BS,
			(float)p.Y * BS + 0.5*BS,
			(float)p.Z * BS + 0.5*BS
		);
	}

	//bool sectorExists(v2s16 p);
	MapSector * getSectorNoGenerate(v2s16 p2d);
	virtual MapSector * emergeSector(v2s16 p);
	
	// Returns InvalidPositionException if not found
	MapBlock * getBlockNoCreate(v3s16 p);
	//virtual MapBlock * getBlock(v3s16 p, bool generate=true);
	
	// Returns InvalidPositionException if not found
	f32 getGroundHeight(v2s16 p, bool generate=false);
	void setGroundHeight(v2s16 p, f32 y, bool generate=false);

	// Returns InvalidPositionException if not found
	bool isNodeUnderground(v3s16 p);
	
	// virtual from NodeContainer
	bool isValidPosition(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock *blockref;
		try{
			blockref = getBlockNoCreate(blockpos);
		}
		catch(InvalidPositionException &e)
		{
			return false;
		}
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		bool is_valid = blockref->isValidPosition(relpos);
		/*std::cout<<"Map::isValidPosition("<<p.X<<","<<p.Y<<","<<p.Z
				<<"): is_valid="<<is_valid<<" (block reported)";*/
		return is_valid;
	}
	
	// virtual from NodeContainer
	MapNode getNode(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock * blockref = getBlockNoCreate(blockpos);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;

		return blockref->getNode(relpos);
	}

	// virtual from NodeContainer
	void setNode(v3s16 p, MapNode & n)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock * blockref = getBlockNoCreate(blockpos);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		blockref->setNode(relpos, n);
	}

	/*MapNode getNodeGenerate(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock * blockref = getBlock(blockpos);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;

		return blockref->getNode(relpos);
	}*/

	/*void setNodeGenerate(v3s16 p, MapNode & n)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock * blockref = getBlock(blockpos);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		blockref->setNode(relpos, n);
	}*/

	void unLightNeighbors(v3s16 pos, u8 oldlight,
			core::list<v3s16> & light_sources,
			core::map<v3s16, MapBlock*> & modified_blocks);
	
	void lightNeighbors(v3s16 pos,
			core::map<v3s16, MapBlock*> & modified_blocks);

	void spreadLight(core::list<v3s16> & from_nodes,
			core::map<v3s16, MapBlock*> & modified_blocks);
	
	v3s16 getBrightestNeighbour(v3s16 p);

	s16 propagateSunlight(v3s16 start,
			core::map<v3s16, MapBlock*> & modified_blocks);
	
	void updateLighting(core::list< MapBlock*>  & a_blocks,
			core::map<v3s16, MapBlock*> & modified_blocks);
			
	void nodeAddedUpdate(v3s16 p, u8 lightwas);
	void removeNodeAndUpdate(v3s16 p);

	core::aabbox3d<s16> getDisplayedBlockArea();

	//bool updateChangedVisibleArea();
	
	void renderMap(video::IVideoDriver* driver,
		video::SMaterial *materials);

	virtual void save(bool only_changed){assert(0);};

	/*
		Updates usage timers
	*/
	void timerUpdate(float dtime);
	
	// Takes cache into account
	void deleteSectors(core::list<v2s16> &list);
	
	// Returns count of deleted sectors
	u32 deleteUnusedSectors(float timeout);
};

class ServerMap : public Map
{
public:
	/*
		savedir: directory to which map data should be saved
	*/
	ServerMap(std::string savedir);
	~ServerMap();

	/*
		Forcefully get a sector from somewhere
	*/
	MapSector * emergeSector(v2s16 p);
	/*
		Forcefully get a block from somewhere
	*/
	MapBlock * emergeBlock(v3s16 p, bool only_from_disk);

	void createDir(std::string path);
	void createSaveDir();
	// returns something like "xxxxxxxx"
	std::string getSectorSubDir(v2s16 pos);
	// returns something like "map/sectors/xxxxxxxx"
	std::string getSectorDir(v2s16 pos);
	std::string createSectorDir(v2s16 pos);
	// dirname: final directory name
	v2s16 getSectorPos(std::string dirname);
	v3s16 getBlockPos(std::string sectordir, std::string blockfile);

	void save(bool only_changed);
	void loadAll();

	void saveMasterHeightmap();
	void loadMasterHeightmap();
	
	// This only saves sector-specific data such as the heightmap
	// (no MapBlocks)
	void saveSectorMeta(MapSector *sector);
	void loadSectorMeta(std::string dirname);
	
	// Full load of a sector including all blocks.
	// returns true on success, false on failure.
	bool loadSectorFull(v2s16 p2d);
	// If sector is not found in memory, try to load it from disk.
	// Returns true if sector now resides in memory
	//bool deFlushSector(v2s16 p2d);

	void saveBlock(MapBlock *block);
	// This will generate a sector with getSector if not found.
	void loadBlock(std::string sectordir, std::string blockfile);

private:
	UnlimitedHeightmap *m_heightmap;
	std::string m_savedir;
	bool m_map_saving_enabled;
};

class Client;

class ClientMap : public Map, public scene::ISceneNode
{
public:
	ClientMap(
			Client *client,
			video::SMaterial *materials,
			scene::ISceneNode* parent,
			scene::ISceneManager* mgr,
			s32 id
	);

	/*
		Forcefully get a sector from somewhere
	*/
	MapSector * emergeSector(v2s16 p);


	/*
		ISceneNode methods
	*/

	virtual void OnRegisterSceneNode()
	{
		if (IsVisible)
			SceneManager->registerNodeForRendering(this);

		ISceneNode::OnRegisterSceneNode();
	}

	virtual void render()
	{
		video::IVideoDriver* driver = SceneManager->getVideoDriver();
		driver->setTransform(video::ETS_WORLD, AbsoluteTransformation);
		renderMap(driver, m_materials);
	}
	
	virtual const core::aabbox3d<f32>& getBoundingBox() const
	{
		return m_box;
	}

	/*virtual u32 getMaterialCount() const
	{
		return 1;
	}

	virtual video::SMaterial& getMaterial(u32 i)
	{
		return materials[0];
	}*/	
	
private:
	Client *m_client;
	
	video::SMaterial *m_materials;

	core::aabbox3d<f32> m_box;
};

#endif

