/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#include "map.h"
//#include "player.h"
#include "main.h"
#include "jmutexautolock.h"
#include "client.h"
#include "filesys.h"
#include "utility.h"

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

/*void * MapUpdateThread::Thread()
{
	if(map == NULL)
		return NULL;
		
	ThreadStarted();

	while(getRun())
	{
		//std::cout<<DTIME<<"UpdateThread running"<<std::endl;
		
		bool did_something;
		did_something = map->updateChangedVisibleArea();
		
		if(did_something == false)
			sleep_ms(500);
	}

	//std::cout<<DTIME<<"UpdateThread stopped"<<std::endl;

	return NULL;
}*/

Map::Map(std::ostream &dout):
	m_dout(dout),
	m_camera_position(0,0,0),
	m_camera_direction(0,0,1),
	//updater(this),
	m_sector_cache(NULL),
	m_hwrapper(this),
	drawoffset(0,0,0)
{
	m_getsector_mutex.Init();
	m_gensector_mutex.Init();
	m_camera_mutex.Init();
	assert(m_getsector_mutex.IsInitialized());
	assert(m_gensector_mutex.IsInitialized());
	assert(m_camera_mutex.IsInitialized());
	
	// Get this so that the player can stay on it at first
	//getSector(v2s16(0,0));
}

Map::~Map()
{
	/*
		Stop updater thread
	*/
	/*updater.setRun(false);
	while(updater.IsRunning())
		sleep_s(1);*/

	/*
		Free all MapSectors.
	*/
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		delete sector;
	}
}

/*
	TODO: These mutexes need rethinking. They don't work properly
	      at the moment. Maybe. Or do they?
*/

/*bool Map::sectorExists(v2s16 p)
{
	JMutexAutoLock lock(m_getsector_mutex);
	core::map<v2s16, MapSector*>::Node *n = m_sectors.find(p);
	return (n != NULL);
}*/

MapSector * Map::getSectorNoGenerate(v2s16 p)
{
	//JMutexAutoLock lock(m_getsector_mutex);

	if(m_sector_cache != NULL && p == m_sector_cache_p){
		MapSector * sector = m_sector_cache;
		// Reset inactivity timer
		sector->usage_timer = 0.0;
		return sector;
	}
	
	core::map<v2s16, MapSector*>::Node *n = m_sectors.find(p);
	// If sector doesn't exist, throw an exception
	if(n == NULL)
	{
		throw InvalidPositionException();
	}
	
	MapSector *sector = n->getValue();
	
	// Cache the last result
	m_sector_cache_p = p;
	m_sector_cache = sector;

	//MapSector * ref(sector);
	
	// Reset inactivity timer
	sector->usage_timer = 0.0;
	return sector;
}

/*
	This is overloaded by ClientMap and ServerMap to allow
	their differing fetch methods.
*/
MapSector * Map::emergeSector(v2s16 p2d)
{
	std::cout<<"Map::emergeSector() stub called"<<std::endl;
	assert(0);
	throw; //vc++ thinks assert = doesn't return a value
}

MapBlock * Map::getBlockNoCreate(v3s16 p3d)
{	
	v2s16 p2d(p3d.X, p3d.Z);
	MapSector * sector = getSectorNoGenerate(p2d);

	return sector->getBlockNoCreate(p3d.Y);
}

/*MapBlock * Map::getBlock(v3s16 p3d, bool generate)
{
	std::cout<<"Map::getBlock() with generate=true called"
			<<std::endl;
	v2s16 p2d(p3d.X, p3d.Z);
	//MapSector * sector = getSector(p2d, generate);
	MapSector * sector = getSectorNoGenerate(p2d);

	if(sector == NULL)
		throw InvalidPositionException();

	return sector->getBlockNoCreate(p3d.Y);
}*/

f32 Map::getGroundHeight(v2s16 p, bool generate)
{
	try{
		v2s16 sectorpos = getNodeSectorPos(p);
		MapSector * sref = getSectorNoGenerate(sectorpos);
		v2s16 relpos = p - sectorpos * MAP_BLOCKSIZE;
		f32 y = sref->getGroundHeight(relpos);
		return y;
	}
	catch(InvalidPositionException &e)
	{
		return GROUNDHEIGHT_NOTFOUND_SETVALUE;
	}
}

void Map::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	/*m_dout<<DTIME<<"Map::setGroundHeight(("
			<<p.X<<","<<p.Y
			<<"), "<<y<<")"<<std::endl;*/
	v2s16 sectorpos = getNodeSectorPos(p);
	MapSector * sref = getSectorNoGenerate(sectorpos);
	v2s16 relpos = p - sectorpos * MAP_BLOCKSIZE;
	//sref->mutex.Lock();
	sref->setGroundHeight(relpos, y);
	//sref->mutex.Unlock();
}

bool Map::isNodeUnderground(v3s16 p)
{
	v3s16 blockpos = getNodeBlockPos(p);
	try{
		MapBlock * block = getBlockNoCreate(blockpos);
		return block->getIsUnderground();
	}
	catch(InvalidPositionException &e)
	{
		return false;
	}
}

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the lighting of
	the node was (before changing it to 0 at the step before), the
	lighting of the neighbour is set to 0 and then the same stuff
	repeats for the neighbour.

	Some things are made strangely to make it as fast as possible.

	Usage: (for clearing all possible spreaded light of a lamp)
		core::list<v3s16> light_sources;
		core::map<v3s16, MapBlock*> modified_blocks;
		u8 oldlight = node_at_pos.light;
		node_at_pos.setLight(0);
		unLightNeighbors(pos, oldlight, light_sources, modified_blocks);
*/
void Map::unLightNeighbors(v3s16 pos, u8 oldlight,
		core::list<v3s16> & light_sources,
		core::map<v3s16, MapBlock*>  & modified_blocks)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	/*
		Initialize block cache
	*/
	v3s16 blockpos_last;
	MapBlock *block = NULL;
	// Cache this a bit, too
	bool block_checked_in_modified = false;
	
	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);

		// Only fetch a new block if the block position has changed
		try{
			if(block == NULL || blockpos != blockpos_last)
			{
				block = getBlockNoCreate(blockpos);
				blockpos_last = blockpos;
				
				block_checked_in_modified = false;
				//blockchangecount++;
			}
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}
		
		// Calculate relative position in block
		v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
		// Get node straight from the block
		MapNode n2 = block->getNode(relpos);
		
		/*
			If the neighbor is dimmer than what was specified
			as oldlight (the light of the previous node)
		*/
		if(n2.getLight() < oldlight)
		{
			/*
				And the neighbor is transparent and it has some light
			*/
			if(n2.transparent() && n2.getLight() != 0)
			{
				/*
					Set light to 0 and recurse.
				*/
				u8 current_light = n2.getLight();
				n2.setLight(0);
				block->setNode(relpos, n2);
				unLightNeighbors(n2pos, current_light,
						light_sources, modified_blocks);
				
				if(block_checked_in_modified == false)
				{
					// If the block is not found in modified_blocks, add.
					if(modified_blocks.find(blockpos) == NULL)
					{
						modified_blocks.insert(blockpos, block);
					}
					block_checked_in_modified = true;
				}
			}
		}
		else{
			light_sources.push_back(n2pos);
		}
	}
}

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the calculated
	lighting coming from the current node to it, the lighting of
	the neighbour is set and the same thing repeats.

	Some things are made strangely to make it as fast as possible.

	Usage:
	core::map<v3s16, MapBlock*> modified_blocks;
	lightNeighbors(pos, node_at_pos.light, modified_blocks);
*/
#if 0
void Map::lightNeighbors(v3s16 pos,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	//std::cout<<"WARNING: Slow Map::lightNeighbors called"<<std::endl;
	
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	core::list<v3s16> neighbour_cache;
	
	/*
		Initialize block cache
		(using center node as a starting position)
	*/
	v3s16 blockpos_last = getNodeBlockPos(pos);
	MapBlock *block = NULL;
	try{
		block = getBlockNoCreate(blockpos_last);
	}
	catch(InvalidPositionException &e)
	{
		return;
	}
	
	// Calculate relative position in block
	v3s16 relpos = pos - blockpos_last * MAP_BLOCKSIZE;
	// Get node straight from the block
	//MapNode *n = block->getNodePtr(relpos);
	MapNode n = block->getNode(relpos);

	u8 oldlight = n.getLight();
	u8 newlight = diminish_light(oldlight);

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);

		try
		{
			/*
				Only fetch a new block if the block position has changed
			*/
			if(blockpos != blockpos_last){
				block = getBlockNoCreate(blockpos);
			}
			blockpos_last = blockpos;

			// Find if block is in cache container
			core::map<v3s16, MapBlock * >::Node *cacheblocknode;
			cacheblocknode = modified_blocks.find(blockpos);
			
			// If the block is not found in the cache
			if(cacheblocknode == NULL)
			{
				/*
					Add block to cache container. It could be a 'set' but
					there is no such thing in Irrlicht.
					
					We can use the pointer as the value of a map just fine,
					it gets nicely cached that way, too.
				*/
				modified_blocks.insert(blockpos, block);
			}
			
			// Calculate relative position in block
			v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
			// Get node straight from the block
			//MapNode *n2 = block->getNodePtr(relpos);
			MapNode n2 = block->getNode(relpos);
			
			/*
				If the neighbor is brighter than the current node,
				add to queue (it will light up this node on its turn)
			*/
			if(n2.light_source() > undiminish_light(oldlight))
			{
				n2.setLight(n2.light_source());
				block->setNode(relpos, n2);
				neighbour_cache.push_back(n2pos);
			}
			/*
				If the neighbor is dimmer than how much light this node
				would spread on it, add to queue
			*/
			if(n2.getLight() < newlight)
			{
				if(n2.transparent())
				{
					n2.setLight(newlight);
					block->setNode(relpos, n2);
					// Cache and recurse at last step for maximum speed
					neighbour_cache.push_back(n2pos);
				}
			}
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}
	}
	
	// Light the neighbors of every block in cache
	core::list<v3s16>::Iterator j = neighbour_cache.begin();
	for(; j != neighbour_cache.end(); j++){
		lightNeighbors(*j, modified_blocks);
	}
}
#endif

void Map::lightNeighbors(v3s16 pos,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	core::list<v3s16> from_nodes;
	from_nodes.push_back(pos);
	spreadLight(from_nodes, modified_blocks);
}

/*
	Lights neighbors of from_nodes, collects all them and then
	goes on recursively.
*/
void Map::spreadLight(core::list<v3s16> & from_nodes,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	const v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};

	if(from_nodes.size() == 0)
		return;
	
	u32 blockchangecount = 0;

	core::list<v3s16> lighted_nodes;
	core::list<v3s16>::Iterator j;
	//j = from_nodes.getIterator();
	j = from_nodes.begin();

	/*
		Initialize block cache
	*/
	v3s16 blockpos_last;
	MapBlock *block = NULL;
	// Cache this a bit, too
	bool block_checked_in_modified = false;
	
	//for(; j.atEnd() == false; j++)
	for(; j != from_nodes.end(); j++)
	{
		//v3s16 pos = j.getNode()->getKey();
		v3s16 pos = *j;
		//std::cout<<"pos=("<<pos.X<<","<<pos.Y<<","<<pos.Z<<")"<<std::endl;
		v3s16 blockpos = getNodeBlockPos(pos);
		
		// Only fetch a new block if the block position has changed
		try{
			if(block == NULL || blockpos != blockpos_last){
				block = getBlockNoCreate(blockpos);
				blockpos_last = blockpos;

				block_checked_in_modified = false;
				blockchangecount++;
			}
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}

		// Calculate relative position in block
		v3s16 relpos = pos - blockpos_last * MAP_BLOCKSIZE;

		// Get node straight from the block
		MapNode n = block->getNode(relpos);

		u8 oldlight = n.getLight();
		u8 newlight = diminish_light(oldlight);

		// Loop through 6 neighbors
		for(u16 i=0; i<6; i++){
			// Get the position of the neighbor node
			v3s16 n2pos = pos + dirs[i];
			
			// Get the block where the node is located
			v3s16 blockpos = getNodeBlockPos(n2pos);

			try
			{
				// Only fetch a new block if the block position has changed
				try{
					if(block == NULL || blockpos != blockpos_last){
						block = getBlockNoCreate(blockpos);
						blockpos_last = blockpos;

						block_checked_in_modified = false;
						blockchangecount++;
					}
				}
				catch(InvalidPositionException &e)
				{
					continue;
				}
				
				// Calculate relative position in block
				v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
				// Get node straight from the block
				MapNode n2 = block->getNode(relpos);
				
				bool changed = false;
				/*
					If the neighbor is brighter than the current node,
					add to list (it will light up this node on its turn)
				*/
				if(n2.getLight() > undiminish_light(oldlight))
				{
					//lighted_nodes.insert(n2pos, true);
					lighted_nodes.push_back(n2pos);
					changed = true;
				}
				/*
					If the neighbor is dimmer than how much light this node
					would spread on it, add to list
				*/
				if(n2.getLight() < newlight)
				{
					if(n2.transparent())
					{
						n2.setLight(newlight);
						block->setNode(relpos, n2);
						lighted_nodes.push_back(n2pos);
						changed = true;
					}
				}

				// Add to modified_blocks
				if(changed == true && block_checked_in_modified == false)
				{
					// If the block is not found in modified_blocks, add.
					if(modified_blocks.find(blockpos) == NULL)
					{
						modified_blocks.insert(blockpos, block);
					}
					block_checked_in_modified = true;
				}
			}
			catch(InvalidPositionException &e)
			{
				continue;
			}
		}
	}

	/*std::cout<<"Changed block "<<blockchangecount<<" times for "
			<<from_nodes.size()<<" nodes"<<std::endl;*/
	
	if(lighted_nodes.size() > 0)
		spreadLight(lighted_nodes, modified_blocks);
}

v3s16 Map::getBrightestNeighbour(v3s16 p)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	u8 brightest_light = 0;
	v3s16 brightest_pos(0,0,0);
	bool found_something = false;

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = p + dirs[i];
		MapNode n2;
		try{
			n2 = getNode(n2pos);
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}
		if(n2.getLight() > brightest_light || found_something == false){
			brightest_light = n2.getLight();
			brightest_pos = n2pos;
			found_something = true;
		}
	}

	if(found_something == false)
		throw InvalidPositionException();
		
	return brightest_pos;
}

/*
	Propagates sunlight down from a node.
	Starting point gets sunlight.

	Returns the lowest y value of where the sunlight went.
*/
s16 Map::propagateSunlight(v3s16 start,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	s16 y = start.Y;
	for(; ; y--)
	{
		v3s16 pos(start.X, y, start.Z);
		
		v3s16 blockpos = getNodeBlockPos(pos);
		MapBlock *block;
		try{
			block = getBlockNoCreate(blockpos);
		}
		catch(InvalidPositionException &e)
		{
			break;
		}

		v3s16 relpos = pos - blockpos*MAP_BLOCKSIZE;
		MapNode n = block->getNode(relpos);

		if(n.transparent())
		{
			n.setLight(LIGHT_SUN);
			block->setNode(relpos, n);

			modified_blocks.insert(blockpos, block);
		}
		else{
			break;
		}
	}
	return y + 1;
}

void Map::updateLighting(core::map<v3s16, MapBlock*> & a_blocks,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	/*m_dout<<DTIME<<"Map::updateLighting(): "
			<<a_blocks.getSize()<<" blocks... ";*/

	//core::list< MapBlock * >::Iterator i = a_blocks.begin();
	
	/*m_dout<<DTIME<<"Map::updateLighting(): "
			"Flooding direct sunlight"<<std::endl;*/
	
	/*
		TODO: Fix lighting

		For now, just dummily light everything
	*/
	/*
	for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;
		v3s16 pos = block->getPos();

		s16 xmin = MAP_BLOCKSIZE*pos.X;
		s16 zmin = MAP_BLOCKSIZE*pos.Z;
		s16 ymin = MAP_BLOCKSIZE*pos.Y;
		s16 xmax = MAP_BLOCKSIZE*pos.X + MAP_BLOCKSIZE - 1;
		s16 zmax = MAP_BLOCKSIZE*pos.Z + MAP_BLOCKSIZE - 1;
		s16 ymax = MAP_BLOCKSIZE*pos.Y + MAP_BLOCKSIZE - 1;

		core::list<v3s16> light_sources;

		for(s16 z=zmin; z<=zmax; z++)
		for(s16 x=xmin; x<=xmax; x++)
		for(s16 y=ymin; y<=ymax; y++)
		{
			v3s16 p(x,y,z);
			MapNode n = getNode(p);
			n.setLight(10);
			setNode(p, n);
		}
	}
	*/

#if 1
	
	// For debugging
	bool debug=false;
	u32 count_was = modified_blocks.size();

	core::map<v3s16, MapBlock*> propagated_blocks;

	/*core::list<MapBlock *>::Iterator i = a_blocks.begin();
	for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;*/

	core::map<v3s16, MapBlock*>::Iterator i;
	i = a_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapBlock *block = i.getNode()->getValue();
		
		for(;;)
		{
			v3s16 pos = block->getPos();
			modified_blocks.insert(pos, block);

			// Unlight neighbours of block
			// This removes spreaded sunlight

			s16 xmin = MAP_BLOCKSIZE*pos.X;
			s16 zmin = MAP_BLOCKSIZE*pos.Z;
			s16 ymin = MAP_BLOCKSIZE*pos.Y;
			s16 xmax = MAP_BLOCKSIZE*pos.X + MAP_BLOCKSIZE - 1;
			s16 zmax = MAP_BLOCKSIZE*pos.Z + MAP_BLOCKSIZE - 1;
			s16 ymax = MAP_BLOCKSIZE*pos.Y + MAP_BLOCKSIZE - 1;

			core::list<v3s16> light_sources;

			/*
				Cache the contents of each neighboring block
			*/

			/*
				Clear all light from center of block
			*/
			for(s16 z=1; z<MAP_BLOCKSIZE-1; z++)
			for(s16 x=1; x<MAP_BLOCKSIZE-1; x++)
			for(s16 y=1; y<MAP_BLOCKSIZE-1; y++)
			{
				
				try{
					v3s16 p(x,y,z);
					MapNode n = block->getNode(v3s16(x,y,z));
					n.setLight(0);
					block->setNode(v3s16(x,y,z), n);
				}
				catch(InvalidPositionException &e)
				{
					std::cout<<"updateLighting(): InvalidPositionException"
							<<std::endl;
				}
			}

			/*
				Unlight all neighbors of block.
				- This will clear all light that has spreaded from
				  this block into others
				- It will also collect all the border nodes from
				  which the light can be re-spreaded
			*/
			for(s16 z=zmin; z<=zmax; z++)
			for(s16 x=xmin; x<=xmax; x++)
			for(s16 y=ymin; y<=ymax; y++)
			{
				// Ignore center nodes
				if(x>=1 && x < MAP_BLOCKSIZE-1
						&& y>=1 && y < MAP_BLOCKSIZE-1
						&& z>=1 && z < MAP_BLOCKSIZE-1)
					continue;

				try{
					// Remove light that has come from this node
					v3s16 p(x,y,z);
					MapNode n = getNode(p);
					u8 oldlight = n.getLight();
					n.setLight(0);
					setNode(p, n);
					unLightNeighbors(p, oldlight,
							light_sources, modified_blocks);
				}
				catch(InvalidPositionException &e)
				{
					std::cout<<"updateLighting(): InvalidPositionException"
							<<std::endl;
				}
			}
			
			if(debug)
			{
				u32 diff = modified_blocks.size() - count_was;
				count_was = modified_blocks.size();
				std::cout<<"unLightNeighbors modified "<<diff<<std::endl;
			}

			bool bottom_valid = block->propagateSunlight(propagated_blocks);

			// TODO: Spread light from propagated sunlight?
			
			/*
				Spread light from collected light sources.
				- They are the nodes in the neighboring blocks that
				  could have lighted up some parts of this block.
			*/
			spreadLight(light_sources, modified_blocks);
			
			if(debug)
			{
				u32 diff = modified_blocks.size() - count_was;
				count_was = modified_blocks.size();
				std::cout<<"spreadLight modified "<<diff<<std::endl;
			}

			// If bottom is valid, we're done.
			if(bottom_valid)
				break;
				
			/*std::cout<<"Bottom for sunlight-propagated block ("
					<<pos.X<<","<<pos.Y<<","<<pos.Z<<") not valid"
					<<std::endl;*/

			// Else get the block below and loop to it

			pos.Y--;
			try{
				block = getBlockNoCreate(pos);
			}
			catch(InvalidPositionException &e)
			{
				assert(0);
			}
			
		}
	}
#endif
#if 0
	/*
		Spread the sunlight
	*/
	
	core::map<v3s16, MapBlock * >::Iterator i2;
	i2 = propagated_blocks.getIterator();
	for(; i2.atEnd() == false; i2++)
	{
		MapBlock *block = i2.getNode()->getValue();
		v3s16 pos = block->getPos();

		/*
			FIXME:
			This spreading of light from neighboring blocks
			messes the lighting up if the neighboring blocks have
			light while they shouldn't

			OK, so the neighboring blocks might have bad lighting.

			-> We'll /unlight/ them!
		*/

		// Go through every node in this block and unlight neighbours
		// Unlight will remove all light that has came from the node.
		
		s16 xmin = MAP_BLOCKSIZE*pos.X;
		s16 zmin = MAP_BLOCKSIZE*pos.Z;
		s16 ymin = MAP_BLOCKSIZE*pos.Y;
		s16 xmax = MAP_BLOCKSIZE*pos.X + MAP_BLOCKSIZE - 1;
		s16 zmax = MAP_BLOCKSIZE*pos.Z + MAP_BLOCKSIZE - 1;
		s16 ymax = MAP_BLOCKSIZE*pos.Y + MAP_BLOCKSIZE - 1;

		core::list<v3s16> light_sources;

		for(s16 z=zmin; z<=zmax; z++)
		for(s16 x=xmin; x<=xmax; x++)
		for(s16 y=ymin; y<=ymax; y++)
		{
			// Remove light that has come from this node
			v3s16 p(x,y,z);
			unLightNeighbors(p, getNode(p).getLight(),
					light_sources, modified_blocks);
		}

		core::list<v3s16>::Iterator j = light_sources.begin();
		for(; j != light_sources.end(); j++)
		{
			lightNeighbors(*j, modified_blocks);
		}

#if 0
		/*
			Now we can spread the light again to this block that
			was coming to this block in the first place.
		*/

		bool changed = false;
		
		if(block->spreadLightIntoBlock())
			changed = true;
		
		if(block->spreadLightInsideBlock())
			changed = true;

		continue;

		/*
			Now spread light to the neighboring blocks
		*/

		/*
			Loop through 3*3*3-1 = 26 neighbors.
			We have to go through them in pretty much
			this exact order (sides first, etc.),
			because sunlight can only go through sides.
		*/

		for(u16 i=0; i<26; i++)
		{
			// Get the position of the neighbor block
			v3s16 pos2 = pos + g_26dirs[i];
			try
			{
				MapBlock *block = getBlockNoCreate(pos2);

				bool changed = false;
				
				// Transfer light from neighbour blocks
				if(block->spreadLightIntoBlock())
					changed = true;
				
				// Transfer light around block
				// (from borders and lamps)
				if(block->spreadLightInsideBlock())
					changed = true;
				
				// This causes some block border faces to be left black?
				// I guess no.
				if(changed)
					modified_blocks.insert(pos2, block);
			}
			catch(InvalidPositionException &e)
			{
				continue;
			}
		}
#endif
	}
#endif
	//m_dout<<"Done ("<<getTimestamp()<<")"<<std::endl;
}

/*
	This is called after changing a node from transparent to opaque.
	The lighting value of the node should be left as-is after changing
	other values. This sets the lighting value to 0.
*/
void Map::nodeAddedUpdate(v3s16 p, u8 lightwas,
		core::map<v3s16, MapBlock*> &modified_blocks)
{
	/*PrintInfo(m_dout);
	m_dout<<DTIME<<"Map::nodeAddedUpdate(): p=("
			<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

	core::list<v3s16> light_sources;
	//MapNode n = getNode(p);

	/*
		From this node to nodes underneath:
		If lighting is sunlight (1.0), unlight neighbours and
		set lighting to 0.
		Else discontinue.
	*/

	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);

	/*
		If there is a node at top and it doesn't have sunlight,
		there has not been any sunlight going down.

		Otherwise there probably is.
	*/
	try{
		MapNode topnode = getNode(toppos);

		if(topnode.getLight() < 0.999)
			node_under_sunlight = false;
	}
	catch(InvalidPositionException &e)
	{
	}

	// Add the block of the added node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock * blockref = getBlockNoCreate(blockpos);
	MapBlock *block = blockref;
	assert(block != NULL);
	modified_blocks.insert(blockpos, block);
	
	// Unlight neighbours of node.
	// This means setting light of all consequent dimmer nodes
	// to 0.
	
	if(isValidPosition(p) == false)
		throw;
		
	unLightNeighbors(p, lightwas, light_sources, modified_blocks);

	MapNode n = getNode(p);
	n.setLight(0);
	setNode(p, n);
	
	if(node_under_sunlight)
	{
		s16 y = p.Y - 1;
		for(;; y--){
			//m_dout<<DTIME<<"y="<<y<<std::endl;
			v3s16 n2pos(p.X, y, p.Z);
			
			MapNode n2;
			try{
				n2 = getNode(n2pos);
			}
			catch(InvalidPositionException &e)
			{
				break;
			}

			if(n2.getLight() >= 0.999){
				//m_dout<<DTIME<<"doing"<<std::endl;
				unLightNeighbors(n2pos, n2.getLight(), light_sources, modified_blocks);
				n2.setLight(0);
				setNode(n2pos, n2);
			}
			else
				break;
		}
	}
	
	core::list<v3s16>::Iterator j = light_sources.begin();
	for(; j != light_sources.end(); j++)
	{
		lightNeighbors(*j, modified_blocks);
	}
}

/*
*/
void Map::removeNodeAndUpdate(v3s16 p,
		core::map<v3s16, MapBlock*> &modified_blocks)
{
	/*PrintInfo(m_dout);
	m_dout<<DTIME<<"Map::removeNodeAndUpdate(): p=("
			<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/
	
	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);
	
	/*
		If there is a node at top and it doesn't have sunlight,
		there will be no sunlight going down.
	*/
	try{
		MapNode topnode = getNode(toppos);

		if(topnode.getLight() != LIGHT_SUN)
			node_under_sunlight = false;
	}
	catch(InvalidPositionException &e)
	{
	}

	/*
		Unlight neighbors (in case the node is a light source)
	*/
	core::list<v3s16> light_sources;
	unLightNeighbors(p, getNode(p).getLight(),
			light_sources, modified_blocks);

	/*
		Remove the node
	*/
	MapNode n;
	n.d = MATERIAL_AIR;
	n.setLight(0);
	setNode(p, n);
	
	/*
		Recalculate lighting
	*/
	core::list<v3s16>::Iterator j = light_sources.begin();
	for(; j != light_sources.end(); j++)
	{
		lightNeighbors(*j, modified_blocks);
	}

	// Add the block of the removed node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock * blockref = getBlockNoCreate(blockpos);
	MapBlock *block = blockref;
	assert(block != NULL);
	modified_blocks.insert(blockpos, block);

	if(node_under_sunlight)
	{
		s16 ybottom = propagateSunlight(p, modified_blocks);
		/*m_dout<<DTIME<<"Node was under sunlight. "
				"Propagating sunlight";
		m_dout<<DTIME<<" -> ybottom="<<ybottom<<std::endl;*/
		s16 y = p.Y;
		for(; y >= ybottom; y--)
		{
			v3s16 p2(p.X, y, p.Z);
			/*m_dout<<DTIME<<"lighting neighbors of node ("
					<<p2.X<<","<<p2.Y<<","<<p2.Z<<")"
					<<std::endl;*/
			lightNeighbors(p2, modified_blocks);
		}
	}
	else
	{
		// Set the lighting of this node to 0
		try{
			MapNode n = getNode(p);
			n.setLight(0);
			setNode(p, n);
		}
		catch(InvalidPositionException &e)
		{
			throw;
		}
	}

	// Get the brightest neighbour node and propagate light from it
	v3s16 n2p = getBrightestNeighbour(p);
	try{
		MapNode n2 = getNode(n2p);
		lightNeighbors(n2p, modified_blocks);
	}
	catch(InvalidPositionException &e)
	{
	}
}

void Map::updateMeshes(v3s16 blockpos)
{
	try{
		v3s16 p = blockpos + v3s16(0,0,0);
		MapBlock *b = getBlockNoCreate(p);
		b->updateMesh();
	}
	catch(InvalidPositionException &e){}
	try{
		v3s16 p = blockpos + v3s16(-1,0,0);
		MapBlock *b = getBlockNoCreate(p);
		b->updateMesh();
	}
	catch(InvalidPositionException &e){}
	try{
		v3s16 p = blockpos + v3s16(0,-1,0);
		MapBlock *b = getBlockNoCreate(p);
		b->updateMesh();
	}
	catch(InvalidPositionException &e){}
	try{
		v3s16 p = blockpos + v3s16(0,0,-1);
		MapBlock *b = getBlockNoCreate(p);
		b->updateMesh();
	}
	catch(InvalidPositionException &e){}
}

void Map::renderMap(video::IVideoDriver* driver,
	video::SMaterial *materials)
{
	//m_dout<<DTIME<<"Rendering map..."<<std::endl;
	
	/*
		Collect all blocks that are in the view range

		Should not optimize more here as we want to auto-update
		all changed nodes in viewing range at the next step.
	*/

	s16 viewing_range_nodes;
	bool viewing_range_all;
	{
		JMutexAutoLock lock(g_range_mutex);
		viewing_range_nodes = g_viewing_range_nodes;
		viewing_range_all = g_viewing_range_all;
	}

	m_camera_mutex.Lock();
	v3f camera_position = m_camera_position;
	v3f camera_direction = m_camera_direction;
	m_camera_mutex.Unlock();

	/*
		Get all blocks and draw all visible ones
	*/

	v3s16 cam_pos_nodes(
			camera_position.X / BS,
			camera_position.Y / BS,
			camera_position.Z / BS);

	v3s16 box_nodes_d = viewing_range_nodes * v3s16(1,1,1);

	v3s16 p_nodes_min = cam_pos_nodes - box_nodes_d;
	v3s16 p_nodes_max = cam_pos_nodes + box_nodes_d;

	// Take a fair amount as we will be dropping more out later
	v3s16 p_blocks_min(
			p_nodes_min.X / MAP_BLOCKSIZE - 1,
			p_nodes_min.Y / MAP_BLOCKSIZE - 1,
			p_nodes_min.Z / MAP_BLOCKSIZE - 1);
	v3s16 p_blocks_max(
			p_nodes_max.X / MAP_BLOCKSIZE + 1,
			p_nodes_max.Y / MAP_BLOCKSIZE + 1,
			p_nodes_max.Z / MAP_BLOCKSIZE + 1);

	core::map<v2s16, MapSector*>::Iterator si;

	si = m_sectors.getIterator();
	for(; si.atEnd() == false; si++)
	{
		MapSector *sector = si.getNode()->getValue();
		v2s16 sp = sector->getPos();
		
		if(viewing_range_all == false)
		{
			if(sp.X < p_blocks_min.X
			|| sp.X > p_blocks_max.X
			|| sp.Y < p_blocks_min.Z
			|| sp.Y > p_blocks_max.Z)
				continue;
		}
		
		core::list< MapBlock * > sectorblocks;
		sector->getBlocks(sectorblocks);
		core::list< MapBlock * >::Iterator i;
		for(i=sectorblocks.begin(); i!=sectorblocks.end(); i++)
		{
			MapBlock *block = *i;

			/*
				Compare block position to camera position, skip
				if not seen on display
			*/
			
			v3s16 blockpos_nodes = block->getPosRelative();

			// Block center position
			v3f blockpos(
					((float)blockpos_nodes.X + MAP_BLOCKSIZE/2) * BS,
					((float)blockpos_nodes.Y + MAP_BLOCKSIZE/2) * BS,
					((float)blockpos_nodes.Z + MAP_BLOCKSIZE/2) * BS
			);

			// Block position relative to camera
			v3f blockpos_relative = blockpos - camera_position;

			// Distance in camera direction (+=front, -=back)
			f32 dforward = blockpos_relative.dotProduct(camera_direction);

			// Total distance
			f32 d = blockpos_relative.getLength();
			
			if(viewing_range_all == false)
			{
				// If block is far away, don't draw it
				if(d > viewing_range_nodes * BS)
					continue;
			}
			
			// Maximum radius of a block
			f32 block_max_radius = 0.5*1.44*1.44*MAP_BLOCKSIZE*BS;
			
			// If block is (nearly) touching the camera, don't
			// bother validating further (that is, render it anyway)
			if(d > block_max_radius * 1.5)
			{
				// Cosine of the angle between the camera direction
				// and the block direction (camera_direction is an unit vector)
				f32 cosangle = dforward / d;
				
				// Compensate for the size of the block
				// (as the block has to be shown even if it's a bit off FOV)
				// This is an estimate.
				cosangle += block_max_radius / dforward;

				// If block is not in the field of view, skip it
				//if(cosangle < cos(FOV_ANGLE/2))
				if(cosangle < cos(FOV_ANGLE/2. * 4./3.))
					continue;
			}
			
			/*
				Draw the faces of the block
			*/
			
			{
				JMutexAutoLock lock(block->mesh_mutex);

				// Cancel if block has no mesh
				if(block->mesh == NULL)
					continue;

				u32 c = block->mesh->getMeshBufferCount();

				for(u32 i=0; i<c; i++)
				{
					scene::IMeshBuffer *buf = block->mesh->getMeshBuffer(i);
					driver->setMaterial(buf->getMaterial());
					driver->drawMeshBuffer(buf);
				}
			}
		}
	}
}

/*
	Updates usage timers
*/
void Map::timerUpdate(float dtime)
{
	core::map<v2s16, MapSector*>::Iterator si;

	si = m_sectors.getIterator();
	for(; si.atEnd() == false; si++)
	{
		MapSector *sector = si.getNode()->getValue();
		sector->usage_timer += dtime;
	}
}

void Map::deleteSectors(core::list<v2s16> &list)
{
	core::list<v2s16>::Iterator j;
	for(j=list.begin(); j!=list.end(); j++)
	{
		MapSector *sector = m_sectors[*j];
		/*
			If sector is in sector cache, remove it from there
		*/
		if(m_sector_cache == sector)
		{
			m_sector_cache = NULL;
		}
		/*
			Remove from map and delete
		*/
		m_sectors.remove(*j);
		delete sector;
	}
}

u32 Map::deleteUnusedSectors(float timeout)
{
	core::list<v2s16> sector_deletion_queue;
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		/*
			Delete sector from memory if it hasn't been used in a long time
		*/
		if(sector->usage_timer > timeout)
		{
			sector_deletion_queue.push_back(i.getNode()->getKey());
		}
	}
	deleteSectors(sector_deletion_queue);
	return sector_deletion_queue.getSize();
}

void Map::PrintInfo(std::ostream &out)
{
	out<<"Map: ";
}

/*
	ServerMap
*/

ServerMap::ServerMap(std::string savedir):
	Map(dout_server),
	m_heightmap(NULL)
{
	m_savedir = savedir;
	m_map_saving_enabled = false;
	
	try
	{
		if(fs::PathExists(m_savedir))
		{
			// Load master heightmap
			loadMasterHeightmap();
			
			// Load sector (0,0) and throw and exception on fail
			if(loadSectorFull(v2s16(0,0)) == false)
				throw LoadError("Failed to load sector (0,0)");

			std::cout<<DTIME<<"Server: Successfully loaded master "
					"heightmap and sector (0,0) from "<<savedir<<
					", assuming valid save directory."
					<<std::endl;

			m_map_saving_enabled = true;
			return;
		}
		else{
			m_map_saving_enabled = true;
		}
	}
	catch(std::exception &e)
	{
		std::cout<<DTIME<<"Server: Failed to load map from "<<savedir
				<<", exception: "<<e.what()<<std::endl;
		std::cout<<DTIME<<"Please remove the map or fix it."<<std::endl;
		std::cout<<DTIME<<"WARNING: Map saving will be disabled."<<std::endl;
	}

	std::cout<<DTIME<<"Initializing new map."<<std::endl;

	//m_heightmap = new UnlimitedHeightmap(16, 42.0, 0.6, 0.0);
	//m_heightmap = new UnlimitedHeightmap(64, 168.0, 0.6, 0.0);
	// This is the latest good one
	m_heightmap = new UnlimitedHeightmap(32, 84.0, 0.6, 0.0);

	// For testing UnlimitedHeightmap
	//m_heightmap = new UnlimitedHeightmap(4, 21.0, 0.6, 0.0);
	//m_heightmap = new UnlimitedHeightmap(8, 60.0, 0.6, 0.0);
	//m_heightmap = new UnlimitedHeightmap(32, 84.0, 0.6, 0.0);

	// Initially write whole map
	save(false);
}

ServerMap::~ServerMap()
{
	try
	{
		if(m_map_saving_enabled)
		{
			//save(false);
			// Save only changed parts
			save(true);
			std::cout<<DTIME<<"Server: saved map to "<<m_savedir<<std::endl;
		}
		else
		{
			std::cout<<DTIME<<"Server: map not saved"<<std::endl;
		}
	}
	catch(std::exception &e)
	{
		std::cout<<DTIME<<"Server: Failed to save map to "<<m_savedir
				<<", exception: "<<e.what()<<std::endl;
	}
	
	if(m_heightmap != NULL)
		delete m_heightmap;
}

MapSector * ServerMap::emergeSector(v2s16 p2d)
{
	// Check that it doesn't exist already
	try{
		return getSectorNoGenerate(p2d);
	}
	catch(InvalidPositionException &e)
	{
	}
	
	/*
		Try to load the sector from disk.
	*/
	if(loadSectorFull(p2d) == true)
	{
		return getSectorNoGenerate(p2d);
	}

	/*
		If there is no heightmap, re-throw the exception.
	*/
	if(m_heightmap == NULL)
	{
		std::cout<<"WARNING: emergeSector(): no heightmap"<<std::endl;
		throw InvalidPositionException("emergeSector(): no heightmap");
	}

	f32 corners[4] = {
		m_heightmap->getGroundHeight(p2d+v2s16(0,0)),
		m_heightmap->getGroundHeight(p2d+v2s16(1,0)),
		m_heightmap->getGroundHeight(p2d+v2s16(1,1)),
		m_heightmap->getGroundHeight(p2d+v2s16(0,1)),
	};
	
	MapSector *sector = new MapSector(this, p2d);

	/*sector->setHeightmap(new FixedHeightmap(&m_hwrapper,
			p2d, MAP_BLOCKSIZE));
	sector->getHeightmap()->generateContinued(2.0, 0.2, corners);*/
	
	/*
		Generate heightmap
	*/
	FixedHeightmap *hm = new FixedHeightmap(&m_hwrapper,
			p2d, MAP_BLOCKSIZE);
	sector->setHeightmap(hm);
	hm->generateContinued(2.0, 0.2, corners);

	/*
		Generate objects
	*/

	float avgheight = (corners[0]+corners[1]+corners[2]+corners[3])/4.0;
	float avgslope = 0.0;
	avgslope += fabs(avgheight - corners[0]);
	avgslope += fabs(avgheight - corners[1]);
	avgslope += fabs(avgheight - corners[2]);
	avgslope += fabs(avgheight - corners[3]);
	avgslope /= 4.0;
	avgslope /= MAP_BLOCKSIZE;
	//std::cout<<"avgslope="<<avgslope<<std::endl;

	float pitness = 0.0;
	v2f32 a;
	a = m_heightmap->getSlope(p2d+v2s16(0,0));
	pitness += -a.X;
	pitness += -a.Y;
	a = m_heightmap->getSlope(p2d+v2s16(0,1));
	pitness += -a.X;
	pitness += a.Y;
	a = m_heightmap->getSlope(p2d+v2s16(1,1));
	pitness += a.X;
	pitness += a.Y;
	a = m_heightmap->getSlope(p2d+v2s16(1,0));
	pitness += a.X;
	pitness += -a.Y;
	pitness /= 4.0;
	pitness /= MAP_BLOCKSIZE;
	
	core::map<v3s16, u8> *objects = new core::map<v3s16, u8>;
	sector->setObjects(objects);

	/*
		Plant some trees if there is not much slope
	*/
	{
		// Avgslope is the derivative of a hill
		u32 tree_max;
		float t = avgslope * avgslope;
		if(t > 0.03)
			tree_max = 15. / (t/0.03);
		else
			tree_max = 15;
		u32 count = (rand()%(tree_max+1));
		for(u32 i=0; i<count; i++)
		{
			s16 x = rand()%(MAP_BLOCKSIZE-2)+1;
			s16 z = rand()%(MAP_BLOCKSIZE-2)+1;
			s16 y = sector->getGroundHeight(v2s16(x,z))+1;
			if(y < WATER_LEVEL)
				continue;
			objects->insert(v3s16(x, y, z),
					HEIGHTMAP_OBJECT_TREE_1);
		}
	}
	{
		// Pitness usually goes at around -0.5...0.5
		//std::cout<<"pitness="<<pitness<<std::endl;
		u32 bush_max = 0;
		if(pitness > 0)
			bush_max = (pitness*100.0);
		if(bush_max > 50)
			bush_max = 50;
		u32 count = (rand()%(bush_max+1));
		for(u32 i=0; i<count; i++)
		{
			s16 x = rand()%(MAP_BLOCKSIZE-0)+0;
			s16 z = rand()%(MAP_BLOCKSIZE-0)+0;
			s16 y = sector->getGroundHeight(v2s16(x,z))+1;
			if(y < WATER_LEVEL)
				continue;
			objects->insert(v3s16(x, y, z),
					HEIGHTMAP_OBJECT_BUSH_1);
		}
	}

	/*
		Insert to container
	*/
	m_sectors.insert(p2d, sector);
	
	return sector;
}

MapBlock * ServerMap::emergeBlock(
		v3s16 p,
		bool only_from_disk,
		core::map<v3s16, MapBlock*> &changed_blocks,
		core::map<v3s16, MapBlock*> &lighting_invalidated_blocks
)
{
	v2s16 p2d(p.X, p.Z);
	s16 block_y = p.Y;
	/*
		This will create or load a sector if not found in memory.
		If block exists on disk, it will be loaded.

		NOTE: On old save formats, this will be slow, as it generates
		      lighting on blocks for them.
	*/
	MapSector *sector = emergeSector(p2d);

	// Try to get a block from the sector
	try{
		MapBlock *block = sector->getBlockNoCreate(block_y);
		return block;
	}
	catch(InvalidPositionException &e)
	{
		// Block not found in sector.
		// If not going to generate a new one, throw.
		if(only_from_disk)
			throw InvalidPositionException
					("emergeBlock(): Not allowed to generate a new block.");
	}

	/*
		OK; Not found.

		Go on generating the block.

		TODO: If a dungeon gets generated so that it's side gets
		      revealed to the outside air, the lighting should be
			  recalculated.
	*/

	MapBlock *block = sector->createBlankBlockNoInsert(block_y);

	// Randomize a bit. This makes dungeons.
	bool low_block_is_empty = false;
	if(rand() % 4 == 0)
		low_block_is_empty = true;
	
	u8 material = MATERIAL_GRASS;
	//s32 avg_ground_y = 0;
	s32 lowest_ground_y = 32767;

	for(s16 z0=0; z0<MAP_BLOCKSIZE; z0++)
	for(s16 x0=0; x0<MAP_BLOCKSIZE; x0++)
	{
		s16 surface_y = sector->getGroundHeight(v2s16(x0,z0));
		//avg_ground_y += surface_y;
		if(surface_y < lowest_ground_y)
			lowest_ground_y = surface_y;

		s32 surface_depth = 0;
		
		float slope = sector->getSlope(v2s16(x0,z0)).getLength();
		
		float min_slope = 0.45;
		float max_slope = 0.85;
		float min_slope_depth = 5.0;
		float max_slope_depth = 0;
		if(slope < min_slope)
			surface_depth = min_slope_depth;
		else if(slope > max_slope)
			surface_depth = max_slope_depth;
		else
			surface_depth = (1.-(slope-min_slope)/max_slope) * min_slope_depth;

		for(s16 y0=0; y0<MAP_BLOCKSIZE; y0++){
			s16 real_y = block_y * MAP_BLOCKSIZE + y0;
			MapNode n;
			/*
				Calculate lighting
				
				FIXME: If there are some man-made structures above the
				newly created block, they won't be taken into account.
			*/
			if(real_y > surface_y)
				n.setLight(LIGHT_SUN);
			/*
				Calculate material
			*/
			// If node is very low
			if(real_y <= surface_y - 10){
				if(low_block_is_empty){
					n.d = MATERIAL_AIR;
				}
				else{
					n.d = MATERIAL_STONE;
				}
			}
			// If node is under surface level
			else if(real_y <= surface_y - surface_depth)
				n.d = MATERIAL_STONE;
			// If node is at or under heightmap y
			else if(real_y <= surface_y)
				n.d = material;
			// If node is over heightmap y
			else{
				// If under water level, it's water
				if(real_y < WATER_LEVEL)
					n.d = MATERIAL_WATER;
				// else air
				else
					n.d = MATERIAL_AIR;
			}
			block->setNode(v3s16(x0,y0,z0), n);
		}
	}

	/*
		Calculate is_underground
	*/
	// Probably underground if the highest part of block is under lowest
	// ground height
	bool is_underground = (block_y+1) * MAP_BLOCKSIZE < lowest_ground_y;
	block->setIsUnderground(is_underground);
	
	/*
		Add block to sector
	*/
	sector->insertBlock(block);
	
	// An y-wise container if changed blocks
	core::map<s16, MapBlock*> changed_blocks_sector;

	/*
		Check if any sector's objects can be placed now.
		If so, place them.
	*/
	core::map<v3s16, u8> *objects = sector->getObjects();
	core::list<v3s16> objects_to_remove;
	for(core::map<v3s16, u8>::Iterator i = objects->getIterator();
			i.atEnd() == false; i++)
	{
		v3s16 p = i.getNode()->getKey();
		u8 d = i.getNode()->getValue();

		//v3s16 p = p_sector - v3s16(0, block_y*MAP_BLOCKSIZE, 0);
		
		if(d == HEIGHTMAP_OBJECT_TEST)
		{
			if(sector->isValidArea(p + v3s16(0,0,0),
					p + v3s16(0,0,0), &changed_blocks_sector))
			{
				MapNode n;
				n.d = MATERIAL_TORCH;
				sector->setNode(p, n);
				objects_to_remove.push_back(p);
			}
		}
		else if(d == HEIGHTMAP_OBJECT_TREE_1)
		{
			v3s16 p_min = p + v3s16(-1,0,-1);
			v3s16 p_max = p + v3s16(1,4,1);
			if(sector->isValidArea(p_min, p_max,
					&changed_blocks_sector))
			{
				MapNode n;
				n.d = MATERIAL_TREE;
				sector->setNode(p+v3s16(0,0,0), n);
				sector->setNode(p+v3s16(0,1,0), n);
				sector->setNode(p+v3s16(0,2,0), n);
				sector->setNode(p+v3s16(0,3,0), n);

				n.d = MATERIAL_LEAVES;

				sector->setNode(p+v3s16(0,4,0), n);
				
				sector->setNode(p+v3s16(-1,4,0), n);
				sector->setNode(p+v3s16(1,4,0), n);
				sector->setNode(p+v3s16(0,4,-1), n);
				sector->setNode(p+v3s16(0,4,1), n);
				sector->setNode(p+v3s16(1,4,1), n);
				sector->setNode(p+v3s16(-1,4,1), n);
				sector->setNode(p+v3s16(-1,4,-1), n);
				sector->setNode(p+v3s16(1,4,-1), n);

				sector->setNode(p+v3s16(-1,3,0), n);
				sector->setNode(p+v3s16(1,3,0), n);
				sector->setNode(p+v3s16(0,3,-1), n);
				sector->setNode(p+v3s16(0,3,1), n);
				sector->setNode(p+v3s16(1,3,1), n);
				sector->setNode(p+v3s16(-1,3,1), n);
				sector->setNode(p+v3s16(-1,3,-1), n);
				sector->setNode(p+v3s16(1,3,-1), n);
				
				objects_to_remove.push_back(p);
				
				// Lighting has to be recalculated for this one.
				sector->getBlocksInArea(p_min, p_max, 
						lighting_invalidated_blocks);
			}
		}
		else if(d == HEIGHTMAP_OBJECT_BUSH_1)
		{
			if(sector->isValidArea(p + v3s16(0,0,0),
					p + v3s16(0,0,0), &changed_blocks_sector))
			{
				MapNode n;
				n.d = MATERIAL_LEAVES;
				sector->setNode(p+v3s16(0,0,0), n);
				
				objects_to_remove.push_back(p);
			}
		}
		else
		{
			std::cerr<<"ServerMap::emergeBlock(): "
					"Invalid heightmap object"
					<<std::endl;
		}
	}

	for(core::list<v3s16>::Iterator i = objects_to_remove.begin();
			i != objects_to_remove.end(); i++)
	{
		objects->remove(*i);
	}

	for(core::map<s16, MapBlock*>::Iterator
			i = changed_blocks_sector.getIterator();
			i.atEnd() == false; i++)
	{
		MapBlock *block = i.getNode()->getValue();

		changed_blocks.insert(block->getPos(), block);
		
		//TODO: REMOVE
		//lighting_invalidated_blocks.insert(block->getPos(), block);
	}

	return block;
}

void ServerMap::createDir(std::string path)
{
	if(fs::CreateDir(path) == false)
	{
		m_dout<<DTIME<<"ServerMap: Failed to create directory "
				<<"\""<<path<<"\""<<std::endl;
		throw BaseException("ServerMap failed to create directory");
	}
}

std::string ServerMap::getSectorSubDir(v2s16 pos)
{
	char cc[9];
	snprintf(cc, 9, "%.4x%.4x",
			(unsigned int)pos.X&0xffff,
			(unsigned int)pos.Y&0xffff);

	return std::string(cc);
}

std::string ServerMap::getSectorDir(v2s16 pos)
{
	return m_savedir + "/sectors/" + getSectorSubDir(pos);
}

v2s16 ServerMap::getSectorPos(std::string dirname)
{
	if(dirname.size() != 8)
		throw InvalidFilenameException("Invalid sector directory name");
	unsigned int x, y;
	int r = sscanf(dirname.c_str(), "%4x%4x", &x, &y);
	if(r != 2)
		throw InvalidFilenameException("Invalid sector directory name");
	v2s16 pos((s16)x, (s16)y);
	return pos;
}

v3s16 ServerMap::getBlockPos(std::string sectordir, std::string blockfile)
{
	v2s16 p2d = getSectorPos(sectordir);

	if(blockfile.size() != 4){
		throw InvalidFilenameException("Invalid block filename");
	}
	unsigned int y;
	int r = sscanf(blockfile.c_str(), "%4x", &y);
	if(r != 1)
		throw InvalidFilenameException("Invalid block filename");
	return v3s16(p2d.X, y, p2d.Y);
}

// Debug helpers
#define ENABLE_SECTOR_SAVING 1
#define ENABLE_SECTOR_LOADING 1
#define ENABLE_BLOCK_SAVING 1
#define ENABLE_BLOCK_LOADING 1

void ServerMap::save(bool only_changed)
{
	if(m_map_saving_enabled == false)
	{
		std::cout<<DTIME<<"WARNING: Not saving map, saving disabled."<<std::endl;
		return;
	}
	
	if(only_changed == false)
		std::cout<<DTIME<<"ServerMap: Saving whole map, this can take time."
				<<std::endl;
	
	saveMasterHeightmap();
	
	u32 sector_meta_count = 0;
	u32 block_count = 0;
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		
		if(ENABLE_SECTOR_SAVING)
		{
			if(sector->differs_from_disk || only_changed == false)
			{
				saveSectorMeta(sector);
				sector_meta_count++;
			}
		}
		if(ENABLE_BLOCK_SAVING)
		{
			core::list<MapBlock*> blocks;
			sector->getBlocks(blocks);
			core::list<MapBlock*>::Iterator j;
			for(j=blocks.begin(); j!=blocks.end(); j++)
			{
				MapBlock *block = *j;
				if(block->getChangedFlag() || only_changed == false)
				{
					saveBlock(block);
					block_count++;
				}
			}
		}
	}
	
	u32 deleted_count = 0;
	deleted_count = deleteUnusedSectors
			(SERVERMAP_DELETE_UNUSED_SECTORS_TIMEOUT);
	
	/*
		Only print if something happened or saved whole map
	*/
	if(only_changed == false || sector_meta_count != 0
			|| block_count != 0 || deleted_count != 0)
	{
		std::cout<<DTIME<<"ServerMap: Written: "
				<<sector_meta_count<<" sector metadata files, "
				<<block_count<<" block files, "
				<<deleted_count<<" sectors unloaded from memory."
				<<std::endl;
	}
}

void ServerMap::loadAll()
{
	std::cout<<DTIME<<"ServerMap: Loading map..."<<std::endl;

	loadMasterHeightmap();

	std::vector<fs::DirListNode> list = fs::GetDirListing(m_savedir+"/sectors/");

	std::cout<<DTIME<<"There are "<<list.size()<<" sectors."<<std::endl;
	
	s32 counter = 0;
	s32 printed_counter = -100000;
	s32 count = list.size();

	std::vector<fs::DirListNode>::iterator i;
	for(i=list.begin(); i!=list.end(); i++)
	{
		if(counter > printed_counter + 10)
		{
			std::cout<<DTIME<<counter<<"/"<<count<<std::endl;
			printed_counter = counter;
		}
		counter++;

		if(ENABLE_SECTOR_LOADING)
		{
			// We want directories
			if(i->dir == false)
				continue;
			try{
				loadSectorMeta(i->name);
			}
			catch(InvalidFilenameException &e)
			{
				// This catches unknown crap in directory
			}
		}
		if(ENABLE_BLOCK_LOADING)
		{
			std::vector<fs::DirListNode> list2 = fs::GetDirListing
					(m_savedir+"/sectors/"+i->name);
			std::vector<fs::DirListNode>::iterator i2;
			for(i2=list2.begin(); i2!=list2.end(); i2++)
			{
				// We want files
				if(i2->dir)
					continue;
				try{
					loadBlock(i->name, i2->name);
				}
				catch(InvalidFilenameException &e)
				{
					// This catches unknown crap in directory
				}
			}
		}
	}
	std::cout<<DTIME<<"ServerMap: Map loaded."<<std::endl;
}

void ServerMap::saveMasterHeightmap()
{
	createDir(m_savedir);
	
	std::string fullpath = m_savedir + "/master_heightmap";
	std::ofstream o(fullpath.c_str(), std::ios_base::binary);
	if(o.good() == false)
		throw FileNotGoodException("Cannot open master heightmap");
	
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;

	SharedBuffer<u8> hmdata = m_heightmap->serialize(version);
	/*
		[0] u8 serialization version
		[1] X master heightmap
	*/
	u32 fullsize = 1 + hmdata.getSize();
	SharedBuffer<u8> data(fullsize);

	data[0] = version;
	memcpy(&data[1], *hmdata, hmdata.getSize());

	o.write((const char*)*data, fullsize);
}

void ServerMap::loadMasterHeightmap()
{
	std::string fullpath = m_savedir + "/master_heightmap";
	std::ifstream is(fullpath.c_str(), std::ios_base::binary);
	if(is.good() == false)
		throw FileNotGoodException("Cannot open master heightmap");
	
	u8 version;
	is.read((char*)&version, 1);

	if(m_heightmap != NULL)
		delete m_heightmap;
	m_heightmap = UnlimitedHeightmap::deSerialize(is, version);
}

void ServerMap::saveSectorMeta(MapSector *sector)
{
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;
	// Get destination
	v2s16 pos = sector->getPos();
	createDir(m_savedir);
	createDir(m_savedir+"/sectors");
	std::string dir = getSectorDir(pos);
	createDir(dir);
	
	std::string fullpath = dir + "/heightmap";
	std::ofstream o(fullpath.c_str(), std::ios_base::binary);
	if(o.good() == false)
		throw FileNotGoodException("Cannot open master heightmap");

	sector->serialize(o, version);
	
	sector->differs_from_disk = false;
}

void ServerMap::loadSectorMeta(std::string dirname)
{
	// Get destination
	v2s16 p2d = getSectorPos(dirname);
	std::string dir = m_savedir + "/sectors/" + dirname;
	
	std::string fullpath = dir + "/heightmap";
	std::ifstream is(fullpath.c_str(), std::ios_base::binary);
	if(is.good() == false)
		throw FileNotGoodException("Cannot open sector heightmap");

	MapSector *sector = MapSector::deSerialize
			(is, this, p2d, &m_hwrapper, m_sectors);
	
	sector->differs_from_disk = false;
}

bool ServerMap::loadSectorFull(v2s16 p2d)
{
	std::string sectorsubdir = getSectorSubDir(p2d);

	if(ENABLE_SECTOR_LOADING)
	{
		try{
			loadSectorMeta(sectorsubdir);
		}
		catch(InvalidFilenameException &e)
		{
			return false;
		}
		catch(FileNotGoodException &e)
		{
			return false;
		}
		catch(std::exception &e)
		{
			return false;
		}
	}
	if(ENABLE_BLOCK_LOADING)
	{
		std::vector<fs::DirListNode> list2 = fs::GetDirListing
				(m_savedir+"/sectors/"+sectorsubdir);
		std::vector<fs::DirListNode>::iterator i2;
		for(i2=list2.begin(); i2!=list2.end(); i2++)
		{
			// We want files
			if(i2->dir)
				continue;
			try{
				loadBlock(sectorsubdir, i2->name);
			}
			catch(InvalidFilenameException &e)
			{
				// This catches unknown crap in directory
			}
		}
	}
	return true;
}

#if 0
bool ServerMap::deFlushSector(v2s16 p2d)
{
	// See if it already exists in memory
	try{
		MapSector *sector = getSectorNoGenerate(p2d);
		return true;
	}
	catch(InvalidPositionException &e)
	{
		/*
			Try to load the sector from disk.
		*/
		if(loadSectorFull(p2d) == true)
		{
			return true;
		}
	}
	return false;
}
#endif

void ServerMap::saveBlock(MapBlock *block)
{
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;
	// Get destination
	v3s16 p3d = block->getPos();
	v2s16 p2d(p3d.X, p3d.Z);
	createDir(m_savedir);
	createDir(m_savedir+"/sectors");
	std::string dir = getSectorDir(p2d);
	createDir(dir);
	
	// Block file is map/sectors/xxxxxxxx/xxxx
	char cc[5];
	snprintf(cc, 5, "%.4x", (unsigned int)p3d.Y&0xffff);
	std::string fullpath = dir + "/" + cc;
	std::ofstream o(fullpath.c_str(), std::ios_base::binary);
	if(o.good() == false)
		throw FileNotGoodException("Cannot open block data");

	/*
		[0] u8 serialization version
		[1] data
	*/
	o.write((char*)&version, 1);
	
	/*u32 block_size =
			MapBlock::serializedLength(version);
	
	SharedBuffer<u8> data(block_size);
	block->serialize(*data, version);*/

	SharedBuffer<u8> data = block->serialize(version);

	o.write((const char*)*data, data.getSize());
	
	// We just wrote it to the disk
	block->resetChangedFlag();
}

void ServerMap::loadBlock(std::string sectordir, std::string blockfile)
{
	// Block file is map/sectors/xxxxxxxx/xxxx
	std::string fullpath = m_savedir+"/sectors/"+sectordir+"/"+blockfile;
	std::ifstream is(fullpath.c_str(), std::ios_base::binary);
	if(is.good() == false)
		throw FileNotGoodException("Cannot open block file");

	v3s16 p3d = getBlockPos(sectordir, blockfile);
	v2s16 p2d(p3d.X, p3d.Z);
	
	u8 version = SER_FMT_VER_INVALID;
	is.read((char*)&version, 1);

	/*u32 block_size = MapBlock::serializedLength(version);
	SharedBuffer<u8> data(block_size);
	is.read((char*)*data, block_size);*/

	// This will always return a sector because we're the server
	MapSector *sector = emergeSector(p2d);
	assert(sector->getPos() == p2d);

	MapBlock *block = NULL;
	bool created_new = false;
	try{
		block = sector->getBlockNoCreate(p3d.Y);
	}
	catch(InvalidPositionException &e)
	{
		block = sector->createBlankBlockNoInsert(p3d.Y);
		created_new = true;
	}

	block->deSerialize(is, version);
	
	if(created_new)
		sector->insertBlock(block);
	
	/*
		Convert old formats to new and save
	*/

	if(version == 0 || version == 1)
	{
		std::cout<<"Block ("<<p3d.X<<","<<p3d.Y<<","<<p3d.Z<<")"
				" is in old format. Updating lighting and saving"
				" modified blocks in new format."<<std::endl;

		// Old version has zero lighting, update it
		core::map<v3s16, MapBlock*> blocks_changed;
		blocks_changed.insert(block->getPos(), block);
		core::map<v3s16, MapBlock*> modified_blocks;
		updateLighting(blocks_changed, modified_blocks);
		
		// Close input file
		is.close();
		
		// Save modified blocks
		core::map<v3s16, MapBlock * >::Iterator i = modified_blocks.getIterator();
		for(; i.atEnd() == false; i++)
		{
			MapBlock *b2 = i.getNode()->getValue();
			saveBlock(b2);
		}
	}
	// Save blocks in new format
	else if(version < SER_FMT_VER_HIGHEST)
	{
		saveBlock(block);
	}
	
	// We just loaded it from the disk, so it's up-to-date.
	block->resetChangedFlag();
}

void ServerMap::PrintInfo(std::ostream &out)
{
	out<<"ServerMap: ";
}

/*
	ClientMap
*/

ClientMap::ClientMap(
		Client *client,
		video::SMaterial *materials,
		scene::ISceneNode* parent,
		scene::ISceneManager* mgr,
		s32 id
):
	Map(dout_client),
	scene::ISceneNode(parent, mgr, id),
	m_client(client),
	m_materials(materials)
{
	/*m_box = core::aabbox3d<f32>(0,0,0,
			map->getW()*BS, map->getH()*BS, map->getD()*BS);*/
	/*m_box = core::aabbox3d<f32>(0,0,0,
			map->getSizeNodes().X * BS,
			map->getSizeNodes().Y * BS,
			map->getSizeNodes().Z * BS);*/
	m_box = core::aabbox3d<f32>(-BS*1000000,-BS*1000000,-BS*1000000,
			BS*1000000,BS*1000000,BS*1000000);
}

MapSector * ClientMap::emergeSector(v2s16 p2d)
{
	// Check that it doesn't exist already
	try{
		return getSectorNoGenerate(p2d);
	}
	catch(InvalidPositionException &e)
	{
	}

	MapSector *sector = new MapSector(this, p2d);

	m_sectors.insert(p2d, sector);

	return sector;
}

void ClientMap::PrintInfo(std::ostream &out)
{
	out<<"ClientMap: ";
}


